# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.cli.commands import _command_init
from kamaki.cli import command
from kamaki.cli.command_tree import CommandTree
from kamaki.cli.errors import CLIError
from kamaki.clients.commissioning import CanonifyException


class commissioning_cli(object):

    api_spec = None
    appname = None
    client = None
    add_context = False
    ctree = None
    description = None

    def __init__(self):
        self.api_spec = self.client.api_spec
        self.appname = self.client.appname
        self.ctree = CommandTree(self.appname, self.description)

    def generate_all(self):
        for f in self.api_spec.call_names():
            c = self.mkClass(f)
            command(self.ctree)(c)

    def mkClass(self, method):
        class C(_command_init):

            __doc__ = self.api_spec.get_doc(method)

            def init(this):
                this.token = (this.config.get(self.appname, 'token') or
                              this.config.get('global', 'token'))
                this.base_url = (this.config.get(self.appname, 'url') or
                                 this.config.get('global', 'url'))
                this.client = self.client(this.base_url, this.token)

            def call(this, method, args):
                ctx = '=null ' if self.add_context else ''
                arglist = '[' + ctx + ' '.join(args) + ']'
                argdict = self.api_spec.parse(method, arglist)
                f = getattr(this.client, method)
                return f(**argdict)

            def main(this, *args):
                this.init()
                try:
                    r = this.call(method, args)
                    print r
                except CanonifyException, e:
                    params = self.api_spec.show_input_canonical(method)
                    meth = method.replace('_', ' ')
                    m = '%s\n  usage: %s %s' % (e, meth, params)
                    raise CLIError('%s: %s\n' % (self.appname, m))
                except Exception as e:
                    raise CLIError('%s: %s\n' % (self.appname, e))

        C.__name__ = self.appname + '_' + method
        return C



#!/usr/bin/env python

# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

import re

keywords = set(['true', 'false', 'null'])
unquoted = set('{}[]"\'0123456789')
name_matcher = re.compile('^[\w @_.+-]+$', re.UNICODE)


def is_name(token):
    if name_matcher.match(token):
        return 1
    return 0


def quote(token, is_dict):
    if not token:
        return '""'

    if not is_name(token[0]):
        comma = ', ' if token[-1] not in '{[' else ''
        return token + comma

    k, sep, v = token.partition('=')
    if not sep or not v.strip('='):
        k, sep, v = token.partition(':')

    if not sep:
        if is_name(token) and token not in keywords:
            token = '"' + token + '"'

        comma = ', ' if token[-1] not in '{[' else ''
        return token + comma

    k = '"' + k + '"'
    is_dict.add(1)

    if not v:
        v = '""'
    else:
        if v.isalnum() and not v.isdigit() and v not in keywords:
            v = '"' + v + '"'

    comma = ', ' if v[-1] not in '{[' else ''
    return k + ':' + v + comma


def clijson(argv):
    tokens = argv
    is_dict = set()

    strlist = []
    append = strlist.append
    dictionary = 0
    token_join = None

    dictionary += 0

    for t in tokens:
        t = t.strip()

        if strlist and t and t in '}]':
            strlist[-1] = strlist[-1].rstrip(', ')

        if token_join:
            t = token_join + t
            token_join = None
        elif t.endswith(':'):
            token_join = t
            continue

        t = quote(t, is_dict)
        append(t)

    if not strlist:
        return 'null'

    if strlist[0][0] not in '{[':
        strlist[-1] = strlist[-1].rstrip(', ')
        o, e = '{}' if is_dict else '[]'
        strlist = [o] + strlist + [e]

    if strlist[-1][-1] in ']}':
        strlist[-1] = strlist[-1].rstrip(',')
    return ''.join(strlist)


if __name__ == '__main__':
    from sys import argv

    print clijson(argv[1:])



# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from imp import find_module, load_module

_modules = {}


def imp_module(fullname):
    if fullname in _modules:
        return _modules[fullname]

    components = fullname.split('.')
    if not components:
        raise ValueError('invalid module name')

    module = None
    modulepath = []

    for name in components:
        if not name:
            raise ValueError("Relative paths not allowed")

        modulepath.append(name)
        modulename = '.'.join(modulepath)
        if modulename in _modules:
            module = _modules[modulename]

        elif hasattr(module, name):
            module = getattr(module, name)

        elif not hasattr(module, '__path__'):
            m = find_module(name)
            module = load_module(modulename, *m)

        else:
            try:
                m = find_module(name, module.__path__)
                module = load_module(modulename, *m)
            except ImportError:
                m = "No module '%s' in '%s'" % (name, module.__path__)
                raise ImportError(m)

        _modules[modulename] = module

    return module


def list_modules():
    return sorted(_modules.keys())



# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, self.list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, self.list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from urlparse import urlparse
from objpool.http import PooledHTTPConnection
from traceback import format_stack

from kamaki.clients.connection import HTTPConnection, HTTPResponse
from kamaki.clients.connection.errors import HTTPConnectionError
from kamaki.clients.connection.errors import HTTPResponseError

from json import loads

from time import sleep
from httplib import ResponseNotReady


class KamakiHTTPResponse(HTTPResponse):
    """The request is created only if there is demand for the response"""

    def __init__(
        self, netloc, scheme,
        method='GET', url='127.0.0.1', headers={}, body=None, poolsize=None):
        self.netloc, self.scheme = netloc, scheme
        self.method, self.url, self.http_headers = method, url, headers
        self.body = body
        self.poolsize = poolsize

    def _get_response(self):
        try:
            if self.prefetched:
                return
        except AttributeError:
            pass

        try:
            with PooledHTTPConnection(
                    self.netloc, self.scheme,
                    size=self.poolsize) as conn:
                super(KamakiHTTPResponse, self).__init__(conn)
                conn.request(
                    method=str(self.method),
                    url=str(self.url),
                    headers=self.http_headers,
                    body=self.body)
                while True:
                    try:
                        r = self.request.getresponse()
                    except ResponseNotReady:
                        sleep(0.001)
                        continue
                    break
                self.prefetched = True
                headers = {}
                for k, v in r.getheaders():
                    headers.update({k: v})
                self.headers = headers
                self.content = r.read()
                self.status_code = r.status
                self.status = r.reason
        except IOError as ioe:
            raise HTTPConnectionError(
                'Cannot connect to %s: %s' % (self.url, ioe.strerror),
                errno=ioe.errno)
        except Exception as err:
            from kamaki.clients import recvlog
            recvlog.debug('\n'.join(['%s' % type(err)] + format_stack()))
            raise HTTPConnectionError(
                'Failed to handle connection to %s %s' % (self.url, err),
                errno=-1)

    @property
    def text(self):
        """
        :returns: (str) content
        """
        self._get_response()
        return '%s' % self._content

    @text.setter
    def text(self, v):
        pass

    @property
    def json(self):
        """
        :returns: (dict) the json-formated content

        :raises HTTPResponseError: if content is not json formated
        """
        self._get_response()
        try:
            return loads(self._content)
        except ValueError as err:
            HTTPResponseError('Response not formated in JSON - %s' % err)

    @json.setter
    def json(self, v):
        pass

    def release(self):
        (self.netloc, self.scheme, self.poolsize) = (None, None, None)


class KamakiHTTPConnection(HTTPConnection):

    def _retrieve_connection_info(self, extra_params={}):
        """
        :param extra_params: (dict) key:val for url parameters

        :returns: (scheme, netloc, url?with&params)
        """
        if self.url:
            url = self.url if self.url[-1] == '/' else (self.url + '/')
        else:
            url = 'http://127.0.0.1'
        if self.path:
            url += self.path[1:] if self.path[0] == '/' else self.path
        params = dict(self.params)
        for k, v in extra_params.items():
            params[k] = v
        for i, (key, val) in enumerate(params.items()):
            param_str = '%s%s' % ('?' if i == 0 else '&', key)
            if val is not None:
                param_str += '=%s' % val
            url += param_str

        parsed = urlparse(url)
        self.url = url
        self.path = parsed.path if parsed.path else '/'
        self.path += '?%s' % parsed.query if parsed.query else ''
        return (parsed.scheme, parsed.netloc)

    def perform_request(
            self,
            method=None, data=None, async_headers={}, async_params={}):
        """
        :param method: (str) http method ('get', 'post', etc.)

        :param data: (binary object)

        :param async_headers: (dict) key:val headers that are used only for one
            request instance as opposed to self.headers, which remain to be
            used by following or parallel requests

        :param async_params: (dict) key:val url parameters that are used only
            for one request instance as opposed to self.params, which remain to
            be used by following or parallel requests

        :returns: (KamakiHTTPResponse) a response object

        :raises HTTPConnectionError: Connection failures
        """
        (scheme, netloc) = self._retrieve_connection_info(
            extra_params=async_params)
        headers = dict(self.headers)
        for k, v in async_headers.items():
            headers[k] = v

        #de-unicode headers to prepare them for http
        http_headers = {}
        for k, v in headers.items():
            http_headers[str(k)] = str(v)

        return KamakiHTTPResponse(
            netloc, scheme, method.upper(), self.path,
            headers=http_headers, body=data, poolsize=self.poolsize)



# Copyright 2012-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients import livetest
from kamaki.clients.astakos import AstakosClient


class Astakos(livetest.Generic):
    def setUp(self):
        self.client = AstakosClient(
            self['astakos', 'url'],
            self['astakos', 'token'])

    def test_authenticate(self):
        self._test_0010_authenticate()

    def _test_0010_authenticate(self):
        r = self.client.authenticate()
        for term in (
                'name',
                'username',
                'auth_token_expires',
                'auth_token_created',
                'uuid',
                'id',
                'email'):
            self.assertTrue(term in r)

    def test_info(self):
        self._test_0020_info()

    def _test_0020_info(self):
        self.assertTrue(set([
            'name',
            'username',
            'uuid']).issubset(self.client.info().keys()))

    def test_get(self):
        self._test_0020_get()

    def _test_0020_get(self):
        for term in ('uuid', 'name', 'username'):
            self.assertEqual(
                self.client.term(term, self['astakos', 'token']),
                self['astakos', term])
        self.assertTrue(self['astakos', 'email'] in self.client.term('email'))

    def test_list(self):
        self.client.authenticate()
        self._test_0020_list()

    def _test_0020_list(self):
        terms = set(['name', 'username', 'uuid', 'email', 'auth_token'])
        uuid = 0
        for r in self.client.list():
            self.assertTrue(terms.issubset(r.keys()))
            self.assertTrue(uuid != r['uuid'] if uuid else True)
            uuid = r['uuid']



# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients import Client, ClientError


class AstakosClient(Client):
    """GRNet Astakos API client"""

    def __init__(self, base_url, token):
        super(AstakosClient, self).__init__(base_url, token)
        self._cache = {}

    def authenticate(self, token=None):
        """Get authentication information and store it in this client
        As long as the AstakosClient instance is alive, the latest
        authentication information for this token will be available

        :param token: (str) custom token to authenticate

        :returns: (dict) authentication information
        """
        self.token = token or self.token
        self._cache[self.token] = self.get('/im/authenticate').json
        return self._cache[self.token]

    def list(self):
        """list cached user information"""
        r = []
        for k, v in self._cache.items():
            r.append(dict(v))
            r[-1].update(dict(auth_token=k))
        return r

    def info(self, token=None):
        """Get (cached) user information"""
        token_bu = self.token
        token = token or self.token
        try:
            r = self._cache[token]
        except KeyError:
            r = self.authenticate(token)
        self.token = token_bu
        return r

    def term(self, key, token=None):
        """Get (cached) term, from user credentials"""
        return self.info(token).get(key, None)



# Copyright 2011-2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERaUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.cli import command
from kamaki.cli.argument import FlagArgument
from kamaki.cli.commands import _command_init, errors
from kamaki.cli.command_tree import CommandTree

config_cmds = CommandTree('config', 'Configuration commands')
_commands = [config_cmds]

about_options = '\nAbout options:\
    \n. syntax: [group.]option\
    \n. example: store.account\
    \n. special case: <option> is equivalent to global.<option>\
    \n. configuration file syntax:\
    \n.   [group]\
    \n.   option=value\
    \n.   (more options can be set per group)'


@command(config_cmds)
class config_list(_command_init):
    """List all configuration options
    FAQ:
    Q: I haven't set any options!
    A: Defaults are used (override with /config set )
    Q: There are more options than I have set
    A: Default options remain if not explicitly replaced or deleted
    """

    @errors.generic.all
    def _run(self):
        for section in sorted(self.config.sections()):
            items = self.config.items(section)
            for key, val in sorted(items):
                print('%s.%s = %s' % (section, key, val))

    def main(self):
        self._run()


@command(config_cmds)
class config_get(_command_init):
    """Show a configuration option"""

    __doc__ += about_options

    @errors.generic.all
    def _run(self, option):
        section, sep, key = option.rpartition('.')
        section = section or 'global'
        value = self.config.get(section, key)
        if value:
            print(value)

    def main(self, option):
        self._run(option)


@command(config_cmds)
class config_set(_command_init):
    """Set a configuration option"""

    __doc__ += about_options

    @errors.generic.all
    def _run(self, option, value):
        section, sep, key = option.rpartition('.')
        section = section or 'global'
        self.config.set(section, key, value)
        self.config.write()
        self.config.reload()

    def main(self, option, value):
        self._run(option, value)


@command(config_cmds)
class config_delete(_command_init):
    """Delete a configuration option
    Default values are not removed by default. To alter this behavior in a
    session, use --default.
    """

    arguments = dict(
        default=FlagArgument(
            'Remove default value as well (persists until end of session)',
            '--default')
    )

    @errors.generic.all
    def _run(self, option):
        section, sep, key = option.rpartition('.')
        section = section or 'global'
        self.config.remove_option(section, key, self['default'])
        self.config.write()
        self.config.reload()

    def main(self, option):
        self._run(option)



# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients.commissioning import Callpoint, CallError
from kamaki.clients.commissioning.utils.debug import debug
from kamaki.clients import Client

from json import loads as json_loads, dumps as json_dumps


class CommissioningClient(Callpoint):

    def __init__(self, base_url, token, poolsize):
        super(CommissioningClient, self).__init__()
        self._kc = Client(base_url, token)
        self._kc.http_client.poolsize = poolsize

    def do_make_call(self, api_call, data):

        _kc = self._kc

        gettable = ['list', 'get', 'read']
        method = (_kc.get if any(api_call.startswith(x) for x in gettable)
                  else _kc.post)

        path = api_call
        json_data = json_dumps(data)
        debug("%s %s\n%s\n<<<\n", method.func_name, path, json_data)

        resp = method(path, data=json_data, success=(200, 450, 500))
        debug(">>>\nStatus: %s", resp.status_code)

        body = resp.text
        debug("\n%s\n<<<\n", body[:128] if body else None)

        status = int(resp.status_code)
        if status == 200:
            return json_loads(body)
        else:
            try:
                error = json_loads(body)
            except ValueError:
                exc = CallError(body, call_error='ValueError')
            else:
                exc = CallError.from_dict(error)
            raise exc



# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.


class Command(object):
    """Store a command and the next-level (2 levels)"""
    _name = None
    path = None
    cmd_class = None
    subcommands = {}
    help = ' '

    def __init__(self, path, help=' ', subcommands={}, cmd_class=None):
        self.path = path
        self.help = help
        self.subcommands = dict(subcommands)
        self.cmd_class = cmd_class

    @property
    def name(self):
        if self._name is None:
            self._name = self.path.split('_')[-1]
        return str(self._name)

    def add_subcmd(self, subcmd):
        if subcmd.path == '%s_%s' % (self.path, subcmd.name):
            self.subcommands[subcmd.name] = subcmd
            return True
        return False

    def get_subcmd(self, name):
        try:
            return self.subcommands[name]
        except KeyError:
            return None

    def contains(self, name):
        """Check if a name is a direct child of self"""
        return name in self.subcommands

    @property
    def is_command(self):
        return self.cmd_class is not None

    @property
    def has_description(self):
        return len(self.help.strip()) > 0

    @property
    def description(self):
        return self.help

    @property
    def parent_path(self):
        parentpath, sep, name = self.path.rpartition('_')
        return parentpath

    def set_class(self, cmd_class):
        self.cmd_class = cmd_class

    def get_class(self):
        return self.cmd_class

    def has_subname(self, subname):
        return subname in self.subcommands

    def get_subnames(self):
        return self.subcommands.keys()

    def get_subcommands(self):
        return self.subcommands.values()

    def sublen(self):
        return len(self.subcommands)

    def parse_out(self, args):
        cmd = self
        index = 0
        for term in args:
            try:
                cmd = cmd.subcommands[term]
            except KeyError:
                break
            index += 1
        return cmd, args[index:]

    def pretty_print(self, recursive=False):
        print('Path: %s (Name: %s) is_cmd: %s\n\thelp: %s' % (
            self.path,
            self.name,
            self.is_command,
            self.help))
        for cmd in self.get_subcommands():
            cmd.pretty_print(recursive)


class CommandTree(object):

    groups = {}
    _all_commands = {}
    name = None
    description = None

    def __init__(self, name, description=''):
        self.name = name
        self.description = description

    def add_command(self, command_path, description=None, cmd_class=None):
        terms = command_path.split('_')
        try:
            cmd = self.groups[terms[0]]
        except KeyError:
            cmd = Command(terms[0])
            self.groups[terms[0]] = cmd
            self._all_commands[terms[0]] = cmd
        path = terms[0]
        for term in terms[1:]:
            path += '_' + term
            try:
                cmd = cmd.subcommands[term]
            except KeyError:
                new_cmd = Command(path)
                self._all_commands[path] = new_cmd
                cmd.add_subcmd(new_cmd)
                cmd = new_cmd
        if cmd_class is not None:
            cmd.set_class(cmd_class)
        if description is not None:
            cmd.help = description

    def find_best_match(self, terms):
        """Find a command that best matches a given list of terms

        :param terms: (list of str) match them against paths in cmd_tree

        :returns: (Command, list) the matching command, the remaining terms
        """
        path = []
        for term in terms:
            check_path = path + [term]
            if '_'.join(check_path) not in self._all_commands:
                break
            path = check_path
        if path:
            return (self._all_commands['_'.join(path)], terms[len(path):])
        return (None, terms)

    def add_tree(self, new_tree):
        tname = new_tree.name
        tdesc = new_tree.description
        self.groups.update(new_tree.groups)
        self._all_commands.update(new_tree._all_commands)
        self.set_description(tname, tdesc)

    def has_command(self, path):
        return path in self._all_commands

    def get_command(self, path):
        return self._all_commands[path]

    def get_groups(self):
        return self.groups.values()

    def get_group_names(self):
        return self.groups.keys()

    def set_description(self, path, description):
        self._all_commands[path].help = description

    def get_description(self, path):
        return self._all_commands[path].help

    def set_class(self, path, cmd_class):
        self._all_commands[path].set_class(cmd_class)

    def get_class(self, path):
        return self._all_commands[path].get_class()

    def get_subnames(self, path=None):
        if path in (None, ''):
            return self.get_group_names()
        return self._all_commands[path].get_subnames()

    def get_subcommands(self, path=None):
        if path in (None, ''):
            return self.get_groups()
        return self._all_commands[path].get_subcommands()

    def get_parent(self, path):
        if '_' not in path:
            return None
        terms = path.split('_')
        parent_path = '_'.join(terms[:-1])
        return self._all_commands[parent_path]

    def get_closest_ancestor_command(self, path):
        path, sep, name = path.rpartition('_')
        while len(path) > 0:
            cmd = self._all_commands[path]
            if cmd.is_command:
                return cmd
            path, sep, name = path.rpartition('_')
        return None

        if '_' not in path:
            return None
        terms = path.split()[:-1]
        while len(terms) > 0:
            tmp_path = '_'.join(terms)
            cmd = self._all_commands[tmp_path]
            if cmd.is_command:
                return cmd
            terms = terms[:-1]
        raise KeyError('No ancestor commands')

    def pretty_print(self, group=None):
        if group is None:
            for group in self.groups:
                self.pretty_print(group)
        else:
            self.groups[group].pretty_print(recursive=True)



# Copyright 2011-2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, self.list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, self.list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from urllib2 import quote
from threading import Thread
from json import dumps, loads
from time import time
import logging
from kamaki.clients.connection.kamakicon import KamakiHTTPConnection
from kamaki.clients.connection.errors import HTTPConnectionError
from kamaki.clients.connection.errors import HTTPResponseError

sendlog = logging.getLogger('clients.send')
datasendlog = logging.getLogger('data.send')
recvlog = logging.getLogger('clients.recv')
datarecvlog = logging.getLogger('data.recv')


class ClientError(Exception):
    def __init__(self, message, status=0, details=None):
        try:
            message += '' if message and message[-1] == '\n' else '\n'
            serv_stat, sep, new_msg = message.partition('{')
            new_msg = sep + new_msg
            json_msg = loads(new_msg)
            key = json_msg.keys()[0]

            json_msg = json_msg[key]
            message = '%s %s (%s)\n' % (serv_stat, key, json_msg['message'])\
                if 'message' in json_msg else '%s %s' % (serv_stat, key)
            if 'code' in json_msg:
                status = json_msg['code']
            if 'details' in json_msg:
                if not details:
                    details = []
                elif not isinstance(details, list):
                    details = [details]
                if json_msg['details']:
                    details.append(json_msg['details'])
        except:
            pass

        super(ClientError, self).__init__(message)
        self.status = status
        self.details = details if details else []


class SilentEvent(Thread):
    """ Thread-run method(*args, **kwargs)
        put exception in exception_bucket
    """
    def __init__(self, method, *args, **kwargs):
        super(self.__class__, self).__init__()
        self.method = method
        self.args = args
        self.kwargs = kwargs

    @property
    def exception(self):
        return getattr(self, '_exception', False)

    @property
    def value(self):
        return getattr(self, '_value', None)

    def run(self):
        try:
            self._value = self.method(*(self.args), **(self.kwargs))
        except Exception as e:
            recvlog.debug('Thread %s got exception %s\n<%s %s' % (
                self,
                type(e),
                e.status if isinstance(e, ClientError) else '',
                e))
            self._exception = e


class Client(object):
    POOL_SIZE = 7

    def __init__(self, base_url, token, http_client=KamakiHTTPConnection()):
        self.base_url = base_url
        self.token = token
        self.headers = {}
        self.DATE_FORMATS = [
            '%a %b %d %H:%M:%S %Y',
            '%A, %d-%b-%y %H:%M:%S GMT',
            '%a, %d %b %Y %H:%M:%S GMT']
        self.http_client = http_client

    def _init_thread_limit(self, limit=1):
        self._thread_limit = limit
        self._elapsed_old = 0.0
        self._elapsed_new = 0.0

    def _watch_thread_limit(self, threadlist):
        recvlog.debug('# running threads: %s' % len(threadlist))
        if (self._elapsed_old > self._elapsed_new) and (
                self._thread_limit < self.POOL_SIZE):
            self._thread_limit += 1
        elif self._elapsed_old < self._elapsed_new and self._thread_limit > 1:
            self._thread_limit -= 1

        self._elapsed_old = self._elapsed_new
        if len(threadlist) >= self._thread_limit:
            self._elapsed_new = 0.0
            for thread in threadlist:
                begin_time = time()
                thread.join()
                self._elapsed_new += time() - begin_time
            self._elapsed_new = self._elapsed_new / len(threadlist)
            return []
        return threadlist

    def _raise_for_status(self, r):
        status_msg = getattr(r, 'status', '')
        try:
            message = '%s %s\n' % (status_msg, r.text)
        except:
            message = '%s %s\n' % (status_msg, r)
        status = getattr(r, 'status_code', getattr(r, 'status', 0))
        raise ClientError(message, status=status)

    def set_header(self, name, value, iff=True):
        """Set a header 'name':'value'"""
        if value is not None and iff:
            self.http_client.set_header(name, value)

    def set_param(self, name, value=None, iff=True):
        if iff:
            self.http_client.set_param(name, value)

    def set_default_header(self, name, value):
        self.http_client.headers.setdefault(name, value)

    def request(
            self,
            method,
            path,
            async_headers={},
            async_params={},
            **kwargs):
        """In threaded/asynchronous requests, headers and params are not safe
        Therefore, the standard self.set_header/param system can be used only
        for headers and params that are common for all requests. All other
        params and headers should passes as
        @param async_headers
        @async_params
        E.g. in most queries the 'X-Auth-Token' header might be the same for
        all, but the 'Range' header might be different from request to request.
        """
        try:
            success = kwargs.pop('success', 200)

            data = kwargs.pop('data', None)
            self.set_default_header('X-Auth-Token', self.token)

            if 'json' in kwargs:
                data = dumps(kwargs.pop('json'))
                self.set_default_header('Content-Type', 'application/json')
            if data:
                self.set_default_header('Content-Length', '%s' % len(data))

            sendlog.info('perform a %s @ %s', method, self.base_url)

            self.http_client.url = self.base_url + (
                '/' if (self.base_url and self.base_url[-1]) != '/' else '')
            self.http_client.path = quote(path.encode('utf8'))
            r = self.http_client.perform_request(
                method,
                data,
                async_headers,
                async_params)

            req = self.http_client
            sendlog.info('%s %s', method, req.url)
            headers = dict(req.headers)
            headers.update(async_headers)

            for key, val in headers.items():
                sendlog.info('\t%s: %s', key, val)
            sendlog.info('')
            if data:
                datasendlog.info(data)

            recvlog.info('%d %s', r.status_code, r.status)
            for key, val in r.headers.items():
                recvlog.info('%s: %s', key, val)
            if r.content:
                datarecvlog.info(r.content)

        except (HTTPResponseError, HTTPConnectionError) as err:
            from traceback import format_stack
            recvlog.debug('\n'.join(['%s' % type(err)] + format_stack()))
            self.http_client.reset_headers()
            self.http_client.reset_params()
            errstr = '%s' % err
            if not errstr:
                errstr = ('%s' % type(err))[7:-2]
            status = getattr(err, 'status', getattr(err, 'errno', 0))
            raise ClientError('%s\n' % errstr, status=status)

        self.http_client.reset_headers()
        self.http_client.reset_params()

        if success is not None:
            # Success can either be an in or a collection
            success = (success,) if isinstance(success, int) else success
            if r.status_code not in success:
                r.release()
                self._raise_for_status(r)
        return r

    def delete(self, path, **kwargs):
        return self.request('delete', path, **kwargs)

    def get(self, path, **kwargs):
        return self.request('get', path, **kwargs)

    def head(self, path, **kwargs):
        return self.request('head', path, **kwargs)

    def post(self, path, **kwargs):
        return self.request('post', path, **kwargs)

    def put(self, path, **kwargs):
        return self.request('put', path, **kwargs)

    def copy(self, path, **kwargs):
        return self.request('copy', path, **kwargs)

    def move(self, path, **kwargs):
        return self.request('move', path, **kwargs)



# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

try:
    from collections import OrderedDict
except ImportError:
    from kamaki.clients.commissioning.utils.ordereddict import OrderedDict
from itertools import chain
from cStringIO import StringIO

ARGMAP_MAGIC = '[=ARGMAP=]'


class arguments(object):
    __slots__ = ('args', 'kw')

    def __init__(self, *args, **kwargs):
        self.args = list(args)
        kw = OrderedDict()
        kwitems = kw.pop('kwitems', None)
        if kwitems is not None:
            kw.update(kwitems)
        kw.update(kwargs)
        self.kw = kw

    def __str__(self):
        return str(self.args) + '+' + str(self.kw)

    def __repr__(self):
        return repr(self.args) + '+' + repr(self.kw)

    def __getitem__(self, key):
        if (isinstance(key, int)) or (
                isinstance(key, long)) or (
                    isinstance(key, slice)):
            return self.args[key]
        else:
            return self.kw[key]

    def __setitem__(self, key, value):
        if (isinstance(key, int)) or (
                isinstance(key, long)) or (
                    isinstance(key, slice)):
            self.args[key] = value
        else:
            self.kw[key] = value

    def __delitem__(self, key):
        if (isinstance(key, int)) or (
                isinstance(key, long)) or (
                    isinstance(key, slice)):
            del self.args[key]
        else:
            del self.kw[key]

    def iteritems(self):
        for item in self.args:
            yield None, item
        for k, v in self.kw:
            yield k, v

    def items(self):
        return list(self.iteritems())

    def iterkeys(self):
        return self.kw.iterkeys()

    def keys(self):
        return self.kw.keys()

    def itervalues(self):
        return chain(self.args, self.kw.itervalues())

    def values(self):
        return list(self.itervalues)

    def append(self, value):
        self.args.append(value)


def argmap_encode(obj, output):
    if obj is None:
        output('[=null]')
        return

    if isinstance(obj, basestring):
        if not obj:
            output('""')
        if isinstance(obj, unicode):
            obj = obj.encode('utf-8')
        output('"')
        start = 0
        while 1:
            end = obj.find(start) + 1
            if end < 0:
                break
            output(obj[start:end] + '"')
            start = end
        output(obj[start:])
        output('"')
        return

    if isinstance(obj, int) or isinstance(obj, long):
        output(str(obj))
        return

    if hasattr(obj, 'iteritems'):
        output('[')
        once = 1
        for k, v in obj.iteritems():
            if once:
                once = 0
            else:
                output(' ')
            if k is not None:
                argmap_encode(k)
                output('=')
            argmap_encode(v)
        output(']')

    if hasattr(obj, '__iter__'):
        output('[')
        once = 1
        for item in obj:
            if once:
                once = 0
            else:
                output(' ')
            argmap_encode(item)
        output(']')

    m = "Unsupported type '%s'" % (type(obj))
    m += ''


def argmap_decode(inputf, s=None):
    if isinstance(inputf, str):
        inputf = StringIO(inputf).read

    if s is None:
        s = inputf(1)

    while 1:
        if not s.isspace():
            break
        s = inputf(1)

    item = ''
    if s == '"':
        s = inputf(1)
        while 1:
            if s == '"':
                s = inputf(1)
                if s != '"':
                    return item, s
            item += s
            s = inputf(1)
    elif s == '[':
        item, s = argmap_decode_args(inputf)
        return item, s
    else:
        while 1:
            item += s
            s = inputf(1)
            if s in ' =]':
                return item, s


def argmap_decode_atom(inputf):
    s = inputf(4)
    if s != 'null':
        m = "Invalid atom '%s'" % (s,)
        raise ValueError(m)
    return None, None


def argmap_decode_args(inputf):
    args = []
    append = args.append
    s = inputf(1)
    key = None

    while 1:
        if s is None:
            s = inputf(1)

        if s == ']':
            if key is not None:
                append((None, key))
            args.append(ARGMAP_MAGIC)
            return args, None

        if s == '=':
            if key is None:
                atom, s = argmap_decode_atom(inputf)
                append((None, atom))
            else:
                value, s = argmap_decode(inputf)
                append((key, value))
                key = None
        elif s == ' ':
            if key is not None:
                append((None, key))
                key = None
            s = inputf(1)
        elif s == '':
            m = "EOF while scanning for ']'"
            raise ValueError(m)
        else:
            if key is not None:
                append((None, key))
            key, s = argmap_decode(inputf, s=s)


def argmap_check(obj):
    if hasattr(obj, 'keys') and callable(obj.keys):
        # this could cover both cases
        return ARGMAP_MAGIC in obj
    if hasattr(obj, '__len__'):
        length = len(obj)
        return length and obj[length - 1] == ARGMAP_MAGIC
    return False


def argmap_unzip_dict(argmap):
    if not hasattr(argmap, 'keys'):
        m = "argmap unzip dict: not a dict"
        raise TypeError(m)
    if ARGMAP_MAGIC not in argmap:
        m = "argmap unzip dict: magic not found"
        raise ValueError(m)
    args = argmap.pop(None, [])
    kw = OrderedDict(argmap)
    del kw[ARGMAP_MAGIC]
    return args, kw


def argmap_unzip_list(argmap):
    if not argmap or argmap.pop() != ARGMAP_MAGIC:
        m = "argmap unzip list: magic not found"
        raise ValueError(m)

    args = []
    append = args.append
    kw = OrderedDict()
    for k, v in argmap:
        if k is None:
            append(v)
        else:
            kw[k] = v
    return args, kw


def argmap_unzip(argmap):
    if hasattr(argmap, 'keys'):
        return argmap_unzip_dict(argmap)
    elif hasattr(argmap, '__iter__'):
        return argmap_unzip_list(argmap)
    else:
        m = "argmap: cannot unzip type %s" % (type(argmap),)
        raise ValueError(m)


def argmap_zip_list(args, kw):
    return [(None, a) for a in args] + kw.items() + [ARGMAP_MAGIC]


def argmap_zip_dict(args, kw):
    argmap = OrderedDict()
    argmap.update(kw)
    argmap[ARGMAP_MAGIC] = ARGMAP_MAGIC
    argmap[None] = list(args) + (argmap[None] if None in argmap else [])
    return argmap

argmap_zip = argmap_zip_list


def argmap_list_to_dict(argmap):
    args, kw = argmap_unzip_list(argmap)
    kw[ARGMAP_MAGIC] = ARGMAP_MAGIC
    kw[None] = args
    return kw


def argmap_dict_to_list(argmap):
    args, kw = argmap_unzip_dict(argmap)
    return args + kw.items() + [ARGMAP_MAGIC]


def argmap_unpack_list(argmap):
    kw = argmap_list_to_dict(argmap)
    if len(kw) == 2:
        return kw[None]
    return kw


def argmap_unpack_dict(argmap):
    if hasattr(argmap, 'keys') and callable(argmap.keys):
        return argmap_dict_to_list(argmap)[:-1]
    return argmap[:-1]



# Copyright 2011 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from kamaki.clients import Client, ClientError
from kamaki.clients.utils import path4url


class ImageClient(Client):
    """OpenStack Image Service API 1.0 and GRNET Plankton client"""

    def __init__(self, base_url, token):
        super(ImageClient, self).__init__(base_url, token)

    def list_public(self, detail=False, filters={}, order=''):
        """
        :param detail: (bool)

        :param filters: (dict) request filters

        :param order: (str) order listing by field (default is ascending, - for
            descending)

        :returns: (list) id,name + full image info if detail
        """
        path = path4url('images', 'detail') if detail else (
            path4url('images') + '/')

        async_params = {}
        if isinstance(filters, dict):
            for key, value in filters.items():
                if value:
                    async_params[key] = value
        if order.startswith('-'):
            self.set_param('sort_dir', 'desc')
            order = order[1:]
        else:
            self.set_param('sort_dir', 'asc')
        self.set_param('sort_key', order, iff=order)

        r = self.get(path, async_params=async_params, success=200)
        return r.json

    def get_meta(self, image_id):
        """
        :param image_id: (string)

        :returns: (list) image metadata (key:val)
        """
        path = path4url('images', image_id)
        r = self.head(path, success=200)

        reply = {}
        properties = {}
        meta_prefix = 'x-image-meta-'
        property_prefix = 'x-image-meta-property-'

        for key, val in r.headers.items():
            key = key.lower()
            if key.startswith(property_prefix):
                key = key[len(property_prefix):]
                properties[key] = val
            elif key.startswith(meta_prefix):
                key = key[len(meta_prefix):]
                reply[key] = val

        if properties:
            reply['properties'] = properties
        return reply

    def register(self, name, location, params={}, properties={}):
        """Register image put at location

        :param name: (str)

        :param location: (str) pithos://<account>/<container>/<path>

        :param params: (dict) image metadata (X-Image-Meta) can be id, store,
            disc_format, container_format, size, checksum, is_public, owner

        :param properties: (dict) image properties (X-Image-Meta-Property)
        """
        path = path4url('images') + '/'
        self.set_header('X-Image-Meta-Name', name)
        self.set_header('X-Image-Meta-Location', location)

        for key, val in params.items():
            if key in ('id', 'store', 'disk_format', 'container_format',
                       'size', 'checksum', 'is_public', 'owner'):
                key = 'x-image-meta-' + key.replace('_', '-')
                self.set_header(key, val)

        for key, val in properties.items():
            self.set_header('X-Image-Meta-Property-%s' % key, val)

        r = self.post(path, success=200)
        r.release()

    def reregister(self, location, name=None, params={}, properties={}):
        """Update existing image (key: location)

        :param location: (str) pithos://<account>/<container>/<path>

        :param name: (str)

        :param params: (dict) image metadata (X-Image-Meta) can be id, store,
            disc_format, container_format, size, checksum, is_public, owner

        :param properties: (dict) image properties (X-Image-Meta-Property)
        """
        path = path4url('images', 'detail')
        r = self.get(path, success=200)
        imgs = [img for img in r.json if img['location'] == location]
        for img in imgs:
            img_name = name if name else img['name']
            img_properties = img['properties']
            for k, v in properties.items():
                img_properties[k] = v
            self.register(img_name, location, params, img_properties)
        r.release()

    def list_members(self, image_id):
        """
        :param image_id: (str)

        :returns: (list) users who can use current user's images
        """
        path = path4url('images', image_id, 'members')
        r = self.get(path, success=200)
        return r.json['members']

    def list_shared(self, member):
        """
        :param member: (str) sharers account

        :returns: (list) images shared by member
        """
        path = path4url('shared-images', member)
        #self.set_param('format', 'json')
        r = self.get(path, success=200)
        return r.json['shared_images']

    def add_member(self, image_id, member):
        """
        :param image_id: (str)

        :param member: (str) user to allow access to current user's images
        """
        path = path4url('images', image_id, 'members', member)
        r = self.put(path, success=204)
        r.release()

    def remove_member(self, image_id, member):
        """
        :param image_id: (str)

        :param member: (str) user to deprive from current user's images
        """
        path = path4url('images', image_id, 'members', member)
        r = self.delete(path, success=204)
        r.release()

    def set_members(self, image_id, members):
        """
        :param image_id: (str)

        :param members: (list) user to deprive from current user's images
        """
        path = path4url('images', image_id, 'members')
        req = {'memberships': [{'member_id': member} for member in members]}
        r = self.put(path, json=req, success=204)
        r.release()



# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.


def str_or_utf8(s):
    if isinstance(s, unicode):
        return s.encode('utf8')
    return str(s)


class CallError(Exception):
    exceptions = {}

    def __new__(cls, *args, **kw):
        call_error = kw.get('call_error', None)
        if call_error is None:
            call_error = cls.__name__
        else:
            call_error = str(call_error)
        cls = CallError.exceptions.get(call_error, cls)
        self = Exception.__new__(cls)
        return self

    def __init__(self, *args, **kw):
        self.call_error = kw.pop('call_error', self.__class__.__name__)
        self.args = args
        self.kwargs = kw

    def __str__(self):
        return '\n--------\n'.join(str_or_utf8(x) for x in self.args)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
                           ','.join(str_or_utf8(x) for x in self.args))

    @classmethod
    def from_exception(cls, exc):
        args = None
        try:
            args = tuple(exc.args)
        except (TypeError, AttributeError):
            pass

        if args is None:
            args = (str(exc),)
        self = cls(*args, call_error=exc.__class__.__name__)
        return self

    def to_dict(self):
        return {'call_error': self.call_error,
                'error_args': (self.args, self.kwargs)}

    @classmethod
    def from_dict(cls, dictobj):
        args = None
        try:
            if 'error_args' in dictobj and 'call_error' in dictobj:
                args = dictobj['error_args']
                call_error = dictobj['call_error']
        except TypeError:
            pass

        if args is None:
            args = (str(dictobj),)
            call_error = 'UnknownError'
            kw = {}
        else:
            args, kw = args

        self = cls(*args, call_error=call_error, **kw)
        return self


def register_exceptions(*exceptions):
    for exception in exceptions:
        if not issubclass(exception, CallError):
            m = "Registering '%s': is not a CallError subclass" % (exception,)
            raise ValueError(m)
        CallError.exceptions[exception.__name__] = exception


def register_exception(exc):
    register_exceptions(exc)
    return exc


@register_exception
class CorruptedError(CallError):
    pass


@register_exception
class InvalidDataError(CallError):
    pass


class ReturnButFail(Exception):
    def __init__(self, retval=None):
        self.data = retval



#!/usr/bin/env python

# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

import logging
from os import environ

_logger = None


def init_logger_file(name, level='DEBUG'):
    logger = logging.getLogger(name)
    handler = logging.FileHandler(name + '.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    level = getattr(logging, level, logging.DEBUG)
    logger.setLevel(level)
    global _logger
    _logger = logger
    return logger


def init_logger_stderr(name, level='DEBUG'):
    logger = logging.getLogger(name)
    from sys import stderr
    handler = logging.StreamHandler(stderr)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    level = getattr(logging, level, logging.DEBUG)
    logger.setLevel(level)
    global _logger
    _logger = logger
    return logger


def debug(fmt, *args):
    global _logger
    if _logger is None:
        init_logger_stderr('logger', get_level())
    _logger.debug(fmt % args)


def get_level(default='INFO'):
    try:
        return environ['DEBUG_LEVEL']
    except:
        return default



# Copyright 2011-2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from sys import stdout
from time import localtime, strftime
from logging import getLogger
from os import path, makedirs

from kamaki.cli import command
from kamaki.cli.command_tree import CommandTree
from kamaki.cli.errors import raiseCLIError, CLISyntaxError
from kamaki.cli.utils import (
    format_size,
    to_bytes,
    print_dict,
    print_items,
    pretty_keys,
    page_hold,
    bold,
    ask_user)
from kamaki.cli.argument import FlagArgument, ValueArgument, IntArgument
from kamaki.cli.argument import KeyValueArgument, DateArgument
from kamaki.cli.argument import ProgressBarArgument
from kamaki.cli.commands import _command_init, errors
from kamaki.clients.pithos import PithosClient, ClientError
from kamaki.clients.astakos import AstakosClient


kloger = getLogger('kamaki')

pithos_cmds = CommandTree('store', 'Pithos+ storage commands')
_commands = [pithos_cmds]


# Argument functionality

class DelimiterArgument(ValueArgument):
    """
    :value type: string
    :value returns: given string or /
    """

    def __init__(self, caller_obj, help='', parsed_name=None, default=None):
        super(DelimiterArgument, self).__init__(help, parsed_name, default)
        self.caller_obj = caller_obj

    @property
    def value(self):
        if self.caller_obj['recursive']:
            return '/'
        return getattr(self, '_value', self.default)

    @value.setter
    def value(self, newvalue):
        self._value = newvalue


class SharingArgument(ValueArgument):
    """Set sharing (read and/or write) groups
    .
    :value type: "read=term1,term2,... write=term1,term2,..."
    .
    :value returns: {'read':['term1', 'term2', ...],
    .   'write':['term1', 'term2', ...]}
    """

    @property
    def value(self):
        return getattr(self, '_value', self.default)

    @value.setter
    def value(self, newvalue):
        perms = {}
        try:
            permlist = newvalue.split(' ')
        except AttributeError:
            return
        for p in permlist:
            try:
                (key, val) = p.split('=')
            except ValueError as err:
                raiseCLIError(
                    err,
                    'Error in --sharing',
                    details='Incorrect format',
                    importance=1)
            if key.lower() not in ('read', 'write'):
                msg = 'Error in --sharing'
                raiseCLIError(err, msg, importance=1, details=[
                    'Invalid permission key %s' % key])
            val_list = val.split(',')
            if not key in perms:
                perms[key] = []
            for item in val_list:
                if item not in perms[key]:
                    perms[key].append(item)
        self._value = perms


class RangeArgument(ValueArgument):
    """
    :value type: string of the form <start>-<end> where <start> and <end> are
        integers
    :value returns: the input string, after type checking <start> and <end>
    """

    @property
    def value(self):
        return getattr(self, '_value', self.default)

    @value.setter
    def value(self, newvalue):
        if newvalue is None:
            self._value = self.default
            return
        (start, end) = newvalue.split('-')
        (start, end) = (int(start), int(end))
        self._value = '%s-%s' % (start, end)

# Command specs


class _pithos_init(_command_init):
    """Initialize a pithos+ kamaki client"""

    @errors.generic.all
    def _run(self):
        self.token = self.config.get('store', 'token')\
            or self.config.get('global', 'token')
        self.base_url = self.config.get('store', 'url')\
            or self.config.get('global', 'url')
        self._set_account()
        self.container = self.config.get('store', 'container')\
            or self.config.get('global', 'container')
        self.client = PithosClient(
            base_url=self.base_url,
            token=self.token,
            account=self.account,
            container=self.container)

    def main(self):
        self._run()

    def _set_account(self):
        astakos = AstakosClient(self.config.get('astakos', 'url'), self.token)
        self.account = self['account'] or astakos.term('uuid')

        """Backwards compatibility"""
        self.account = self.account\
            or self.config.get('store', 'account')\
            or self.config.get('global', 'account')


class _store_account_command(_pithos_init):
    """Base class for account level storage commands"""

    def __init__(self, arguments={}):
        super(_store_account_command, self).__init__(arguments)
        self['account'] = ValueArgument(
            'Set user account (not permanent)',
            '--account')

    def _run(self):
        super(_store_account_command, self)._run()
        if self['account']:
            self.client.account = self['account']

    @errors.generic.all
    def main(self):
        self._run()


class _store_container_command(_store_account_command):
    """Base class for container level storage commands"""

    container = None
    path = None

    def __init__(self, arguments={}):
        super(_store_container_command, self).__init__(arguments)
        self['container'] = ValueArgument(
            'Set container to work with (temporary)',
            '--container')

    @errors.generic.all
    def _dest_container_path(self, dest_container_path):
        if self['destination_container']:
            return (self['destination_container'], dest_container_path)
        dst = dest_container_path.split(':')
        return (dst[0], dst[1]) if len(dst) > 1 else (None, dst[0])

    def extract_container_and_path(
            self,
            container_with_path,
            path_is_optional=True):
        """Contains all heuristics for deciding what should be used as
        container or path. Options are:
        * user string of the form container:path
        * self.container, self.path variables set by super constructor, or
        explicitly by the caller application
        Error handling is explicit as these error cases happen only here
        """
        try:
            assert isinstance(container_with_path, str)
        except AssertionError as err:
            if self['container'] and path_is_optional:
                self.container = self['container']
                self.client.container = self['container']
                return
            raiseCLIError(err)

        user_cont, sep, userpath = container_with_path.partition(':')

        if sep:
            if not user_cont:
                raiseCLIError(CLISyntaxError(
                    'Container is missing\n',
                    details=errors.pithos.container_howto))
            alt_cont = self['container']
            if alt_cont and user_cont != alt_cont:
                raiseCLIError(CLISyntaxError(
                    'Conflict: 2 containers (%s, %s)' % (user_cont, alt_cont),
                    details=errors.pithos.container_howto)
                )
            self.container = user_cont
            if not userpath:
                raiseCLIError(CLISyntaxError(
                    'Path is missing for object in container %s' % user_cont,
                    details=errors.pithos.container_howto)
                )
            self.path = userpath
        else:
            alt_cont = self['container'] or self.client.container
            if alt_cont:
                self.container = alt_cont
                self.path = user_cont
            elif path_is_optional:
                self.container = user_cont
                self.path = None
            else:
                self.container = user_cont
                raiseCLIError(CLISyntaxError(
                    'Both container and path are required',
                    details=errors.pithos.container_howto)
                )

    @errors.generic.all
    def _run(self, container_with_path=None, path_is_optional=True):
        super(_store_container_command, self)._run()
        if self['container']:
            self.client.container = self['container']
            if container_with_path:
                self.path = container_with_path
            elif not path_is_optional:
                raise CLISyntaxError(
                    'Both container and path are required',
                    details=errors.pithos.container_howto)
        elif container_with_path:
            self.extract_container_and_path(
                container_with_path,
                path_is_optional)
            self.client.container = self.container
        self.container = self.client.container

    def main(self, container_with_path=None, path_is_optional=True):
        self._run(container_with_path, path_is_optional)


@command(pithos_cmds)
class store_list(_store_container_command):
    """List containers, object trees or objects in a directory
    Use with:
    1 no parameters : containers in current account
    2. one parameter (container) or --container : contents of container
    3. <container>:<prefix> or --container=<container> <prefix>: objects in
    .   container starting with prefix
    """

    arguments = dict(
        detail=FlagArgument('show detailed output', '-l'),
        limit=IntArgument('limit the number of listed items', '-n'),
        marker=ValueArgument('show output greater that marker', '--marker'),
        prefix=ValueArgument('show output starting with prefix', '--prefix'),
        delimiter=ValueArgument('show output up to delimiter', '--delimiter'),
        path=ValueArgument(
            'show output starting with prefix up to /',
            '--path'),
        meta=ValueArgument(
            'show output with specified meta keys',
            '--meta',
            default=[]),
        if_modified_since=ValueArgument(
            'show output modified since then',
            '--if-modified-since'),
        if_unmodified_since=ValueArgument(
            'show output not modified since then',
            '--if-unmodified-since'),
        until=DateArgument('show metadata until then', '--until'),
        format=ValueArgument(
            'format to parse until data (default: d/m/Y H:M:S )',
            '--format'),
        shared=FlagArgument('show only shared', '--shared'),
        public=FlagArgument('show only public', '--public'),
        more=FlagArgument(
            'output results in pages (-n to set items per page, default 10)',
            '--more'),
        exact_match=FlagArgument(
            'Show only objects that match exactly with path',
            '--exact-match')
    )

    def print_objects(self, object_list):
        limit = int(self['limit']) if self['limit'] > 0 else len(object_list)
        for index, obj in enumerate(object_list):
            if self['exact_match'] and self.path and not (
                    obj['name'] == self.path or 'content_type' in obj):
                continue
            pretty_obj = obj.copy()
            index += 1
            empty_space = ' ' * (len(str(len(object_list))) - len(str(index)))
            if obj['content_type'] == 'application/directory':
                isDir = True
                size = 'D'
            else:
                isDir = False
                size = format_size(obj['bytes'])
                pretty_obj['bytes'] = '%s (%s)' % (obj['bytes'], size)
            oname = bold(obj['name'])
            if self['detail']:
                print('%s%s. %s' % (empty_space, index, oname))
                print_dict(pretty_keys(pretty_obj), exclude=('name'))
                print
            else:
                oname = '%s%s. %6s %s' % (empty_space, index, size, oname)
                oname += '/' if isDir else ''
                print(oname)
            if self['more']:
                page_hold(index, limit, len(object_list))

    def print_containers(self, container_list):
        limit = int(self['limit']) if self['limit'] > 0\
            else len(container_list)
        for index, container in enumerate(container_list):
            if 'bytes' in container:
                size = format_size(container['bytes'])
            cname = '%s. %s' % (index + 1, bold(container['name']))
            if self['detail']:
                print(cname)
                pretty_c = container.copy()
                if 'bytes' in container:
                    pretty_c['bytes'] = '%s (%s)' % (container['bytes'], size)
                print_dict(pretty_keys(pretty_c), exclude=('name'))
                print
            else:
                if 'count' in container and 'bytes' in container:
                    print('%s (%s, %s objects)' % (
                        cname,
                        size,
                        container['count']))
                else:
                    print(cname)
            if self['more']:
                page_hold(index + 1, limit, len(container_list))

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.object_path
    @errors.pithos.container
    def _run(self):
        if self.container is None:
            r = self.client.account_get(
                limit=False if self['more'] else self['limit'],
                marker=self['marker'],
                if_modified_since=self['if_modified_since'],
                if_unmodified_since=self['if_unmodified_since'],
                until=self['until'],
                show_only_shared=self['shared'])
            self.print_containers(r.json)
        else:
            prefix = self.path or self['prefix']
            r = self.client.container_get(
                limit=False if self['more'] else self['limit'],
                marker=self['marker'],
                prefix=prefix,
                delimiter=self['delimiter'],
                path=self['path'],
                if_modified_since=self['if_modified_since'],
                if_unmodified_since=self['if_unmodified_since'],
                until=self['until'],
                meta=self['meta'],
                show_only_shared=self['shared'])
            self.print_objects(r.json)

    def main(self, container____path__=None):
        super(self.__class__, self)._run(container____path__)
        self._run()


@command(pithos_cmds)
class store_mkdir(_store_container_command):
    """Create a directory"""

    __doc__ += '\n. '.join([
        'Kamaki hanldes directories the same way as OOS Storage and Pithos+:',
        'A   directory  is   an  object  with  type  "application/directory"',
        'An object with path  dir/name can exist even if  dir does not exist',
        'or even if dir  is  a non  directory  object.  Users can modify dir',
        'without affecting the dir/name object in any way.'])

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self):
        self.client.create_directory(self.path)

    def main(self, container___directory):
        super(self.__class__, self)._run(
            container___directory,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_touch(_store_container_command):
    """Create an empty object (file)
    If object exists, this command will reset it to 0 length
    """

    arguments = dict(
        content_type=ValueArgument(
            'Set content type (default: application/octet-stream)',
            '--content-type',
            default='application/octet-stream')
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self):
        self.client.create_object(self.path, self['content_type'])

    def main(self, container___path):
        super(store_touch, self)._run(
            container___path,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_create(_store_container_command):
    """Create a container"""

    arguments = dict(
        versioning=ValueArgument(
            'set container versioning (auto/none)',
            '--versioning'),
        quota=IntArgument('set default container quota', '--quota'),
        meta=KeyValueArgument(
            'set container metadata (can be repeated)',
            '--meta')
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self):
        self.client.container_put(
            quota=self['quota'],
            versioning=self['versioning'],
            metadata=self['meta'])

    def main(self, container=None):
        super(self.__class__, self)._run(container)
        if container and self.container != container:
            raiseCLIError('Invalid container name %s' % container, details=[
                'Did you mean "%s" ?' % self.container,
                'Use --container for names containing :'])
        self._run()


@command(pithos_cmds)
class store_copy(_store_container_command):
    """Copy objects from container to (another) container
    Semantics:
    copy cont:path path2
    .   will copy all <obj> prefixed with path, as path2<obj>
    .   or as path2 if path corresponds to just one whole object
    copy cont:path cont2:
    .   will copy all <obj> prefixed with path to container cont2
    copy cont:path [cont2:]path2 --exact-match
    .   will copy at most one <obj> as a new object named path2,
    .   provided path corresponds to a whole object path
    copy cont:path [cont2:]path2 --replace
    .   will copy all <obj> prefixed with path, replacing path with path2
    where <obj> is a full file or directory object path.
    Use options:
    1. <container1>:<path1> [container2:]<path2> : if container2 is not given,
    destination is container1:path2
    2. <container>:<path1> <path2> : make a copy in the same container
    3. Can use --container= instead of <container1>
    """

    arguments = dict(
        destination_container=ValueArgument(
            'use it if destination container name contains a : character',
            '--dst-container'),
        source_version=ValueArgument(
            'copy specific version',
            '--source-version'),
        public=ValueArgument('make object publicly accessible', '--public'),
        content_type=ValueArgument(
            'change object\'s content type',
            '--content-type'),
        recursive=FlagArgument(
            'mass copy with delimiter /',
            ('-r', '--recursive')),
        exact_match=FlagArgument(
            'Copy only the object that fully matches path',
            '--exact-match'),
        replace=FlagArgument('Replace src. path with dst. path', '--replace')
    )

    def _objlist(self, dst_path):
        if self['exact_match']:
            return [(dst_path or self.path, self.path)]
        r = self.client.container_get(prefix=self.path)
        if len(r.json) == 1:
            obj = r.json[0]
            return [(obj['name'], dst_path or obj['name'])]
        start = len(self.path) if self['replace'] else 0
        return [(obj['name'], '%s%s' % (
            dst_path,
            obj['name'][start:])) for obj in r.json]

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self, dst_cont, dst_path):
        no_source_object = True
        for src_object, dst_object in self._objlist(dst_path):
            no_source_object = False
            self.client.copy_object(
                src_container=self.container,
                src_object=src_object,
                dst_container=dst_cont or self.container,
                dst_object=dst_object,
                source_version=self['source_version'],
                public=self['public'],
                content_type=self['content_type'])
        if no_source_object:
            raiseCLIError('No object %s in container %s' % (
                self.path,
                self.container))

    def main(
            self,
            source_container___path,
            destination_container___path=None):
        super(self.__class__, self)._run(
            source_container___path,
            path_is_optional=False)
        (dst_cont, dst_path) = self._dest_container_path(
            destination_container___path)
        self._run(dst_cont=dst_cont, dst_path=dst_path or '')


@command(pithos_cmds)
class store_move(_store_container_command):
    """Move/rename objects
    Semantics:
    move cont:path path2
    .   will move all <obj> prefixed with path, as path2<obj>
    .   or as path2 if path corresponds to just one whole object
    move cont:path cont2:
    .   will move all <obj> prefixed with path to container cont2
    move cont:path [cont2:]path2 --exact-match
    .   will move at most one <obj> as a new object named path2,
    .   provided path corresponds to a whole object path
    move cont:path [cont2:]path2 --replace
    .   will move all <obj> prefixed with path, replacing path with path2
    where <obj> is a full file or directory object path.
    Use options:
    1. <container1>:<path1> [container2:]<path2> : if container2 not given,
    destination is container1:path2
    2. <container>:<path1> path2 : rename
    3. Can use --container= instead of <container1>
    """

    arguments = dict(
        destination_container=ValueArgument(
            'use it if destination container name contains a : character',
            '--dst-container'),
        source_version=ValueArgument('specify version', '--source-version'),
        public=FlagArgument('make object publicly accessible', '--public'),
        content_type=ValueArgument('modify content type', '--content-type'),
        recursive=FlagArgument('up to delimiter /', ('-r', '--recursive')),
        exact_match=FlagArgument(
            'Copy only the object that fully matches path',
            '--exact-match'),
        replace=FlagArgument('Replace src. path with dst. path', '--replace')
    )

    def _objlist(self, dst_path):
        if self['exact_match']:
            return [(dst_path or self.path, self.path)]
        r = self.client.container_get(prefix=self.path)
        if len(r.json) == 1:
            obj = r.json[0]
            return [(obj['name'], dst_path or obj['name'])]
        return [(
            obj['name'],
            '%s%s' % (
                dst_path,
                obj['name'][len(self.path) if self['replace'] else 0:]
            )) for obj in r.json]

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self, dst_cont, dst_path):
        no_source_object = True
        for src_object, dst_object in self._objlist(dst_path):
            no_source_object = False
            self.client.move_object(
                src_container=self.container,
                src_object=src_object,
                dst_container=dst_cont or self.container,
                dst_object=dst_object,
                source_version=self['source_version'],
                public=self['public'],
                content_type=self['content_type'])
        if no_source_object:
            raiseCLIError('No object %s in container %s' % (
                self.path,
                self.container))

    def main(
            self,
            source_container___path,
            destination_container___path=None):
        super(self.__class__, self)._run(
            source_container___path,
            path_is_optional=False)
        (dst_cont, dst_path) = self._dest_container_path(
            destination_container___path)
        self._run(dst_cont=dst_cont, dst_path=dst_path or '')


@command(pithos_cmds)
class store_append(_store_container_command):
    """Append local file to (existing) remote object
    The remote object should exist.
    If the remote object is a directory, it is transformed into a file.
    In the later case, objects under the directory remain intact.
    """

    arguments = dict(
        progress_bar=ProgressBarArgument(
            'do not show progress bar',
            '--no-progress-bar',
            default=False)
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self, local_path):
        (progress_bar, upload_cb) = self._safe_progress_bar('Appending')
        try:
            f = open(local_path, 'rb')
            self.client.append_object(self.path, f, upload_cb)
        except Exception:
            self._safe_progress_bar_finish(progress_bar)
            raise
        finally:
            self._safe_progress_bar_finish(progress_bar)

    def main(self, local_path, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self._run(local_path)


@command(pithos_cmds)
class store_truncate(_store_container_command):
    """Truncate remote file up to a size (default is 0)"""

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    @errors.pithos.object_size
    def _run(self, size=0):
        self.client.truncate_object(self.path, size)

    def main(self, container___path, size=0):
        super(self.__class__, self)._run(container___path)
        self._run(size=size)


@command(pithos_cmds)
class store_overwrite(_store_container_command):
    """Overwrite part (from start to end) of a remote file
    overwrite local-path container 10 20
    .   will overwrite bytes from 10 to 20 of a remote file with the same name
    .   as local-path basename
    overwrite local-path container:path 10 20
    .   will overwrite as above, but the remote file is named path
    """

    arguments = dict(
        progress_bar=ProgressBarArgument(
            'do not show progress bar',
            '--no-progress-bar',
            default=False)
    )

    def _open_file(self, local_path, start):
        f = open(path.abspath(local_path), 'rb')
        f.seek(0, 2)
        f_size = f.tell()
        f.seek(start, 0)
        return (f, f_size)

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    @errors.pithos.object_size
    def _run(self, local_path, start, end):
        (start, end) = (int(start), int(end))
        (f, f_size) = self._open_file(local_path, start)
        (progress_bar, upload_cb) = self._safe_progress_bar(
            'Overwrite %s bytes' % (end - start))
        try:
            self.client.overwrite_object(
                obj=self.path,
                start=start,
                end=end,
                source_file=f,
                upload_cb=upload_cb)
        except Exception:
            self._safe_progress_bar_finish(progress_bar)
            raise
        finally:
            self._safe_progress_bar_finish(progress_bar)

    def main(self, local_path, container___path, start, end):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=None)
        self.path = self.path or path.basename(local_path)
        self._run(local_path=local_path, start=start, end=end)


@command(pithos_cmds)
class store_manifest(_store_container_command):
    """Create a remote file of uploaded parts by manifestation
    Remains functional for compatibility with OOS Storage. Users are advised
    to use the upload command instead.
    Manifestation is a compliant process for uploading large files. The files
    have to be chunked in smalled files and uploaded as <prefix><increment>
    where increment is 1, 2, ...
    Finally, the manifest command glues partial files together in one file
    named <prefix>
    The upload command is faster, easier and more intuitive than manifest
    """

    arguments = dict(
        etag=ValueArgument('check written data', '--etag'),
        content_encoding=ValueArgument(
            'set MIME content type',
            '--content-encoding'),
        content_disposition=ValueArgument(
            'the presentation style of the object',
            '--content-disposition'),
        content_type=ValueArgument(
            'specify content type',
            '--content-type',
            default='application/octet-stream'),
        sharing=SharingArgument(
            '\n'.join([
                'define object sharing policy',
                '    ( "read=user1,grp1,user2,... write=user1,grp2,..." )']),
            '--sharing'),
        public=FlagArgument('make object publicly accessible', '--public')
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        self.client.create_object_by_manifestation(
            self.path,
            content_encoding=self['content_encoding'],
            content_disposition=self['content_disposition'],
            content_type=self['content_type'],
            sharing=self['sharing'],
            public=self['public'])

    def main(self, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self.run()


@command(pithos_cmds)
class store_upload(_store_container_command):
    """Upload a file"""

    arguments = dict(
        use_hashes=FlagArgument(
            'provide hashmap file instead of data',
            '--use-hashes'),
        etag=ValueArgument('check written data', '--etag'),
        unchunked=FlagArgument('avoid chunked transfer mode', '--unchunked'),
        content_encoding=ValueArgument(
            'set MIME content type',
            '--content-encoding'),
        content_disposition=ValueArgument(
            'specify objects presentation style',
            '--content-disposition'),
        content_type=ValueArgument('specify content type', '--content-type'),
        sharing=SharingArgument(
            help='\n'.join([
                'define sharing object policy',
                '( "read=user1,grp1,user2,... write=user1,grp2,... )']),
            parsed_name='--sharing'),
        public=FlagArgument('make object publicly accessible', '--public'),
        poolsize=IntArgument('set pool size', '--with-pool-size'),
        progress_bar=ProgressBarArgument(
            'do not show progress bar',
            '--no-progress-bar',
            default=False),
        overwrite=FlagArgument('Force overwrite, if object exists', '-f')
    )

    def _remote_path(self, remote_path, local_path=''):
        if self['overwrite']:
            return remote_path
        try:
            r = self.client.get_object_info(remote_path)
        except ClientError as ce:
            if ce.status == 404:
                return remote_path
            raise ce
        ctype = r.get('content-type', '')
        if 'application/directory' == ctype.lower():
            ret = '%s/%s' % (remote_path, local_path)
            return self._remote_path(ret) if local_path else ret
        raiseCLIError(
            'Object %s already exists' % remote_path,
            importance=1,
            details=['use -f to overwrite or resume'])

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    @errors.pithos.local_path
    def _run(self, local_path, remote_path):
        poolsize = self['poolsize']
        if poolsize > 0:
            self.client.POOL_SIZE = int(poolsize)
        params = dict(
            content_encoding=self['content_encoding'],
            content_type=self['content_type'],
            content_disposition=self['content_disposition'],
            sharing=self['sharing'],
            public=self['public'])
        remote_path = self._remote_path(remote_path, local_path)
        with open(path.abspath(local_path), 'rb') as f:
            if self['unchunked']:
                self.client.upload_object_unchunked(
                    remote_path,
                    f,
                    etag=self['etag'],
                    withHashFile=self['use_hashes'],
                    **params)
            else:
                try:
                    (progress_bar, upload_cb) = self._safe_progress_bar(
                        'Uploading')
                    if progress_bar:
                        hash_bar = progress_bar.clone()
                        hash_cb = hash_bar.get_generator(
                            'Calculating block hashes'
                        )
                    else:
                        hash_cb = None
                    self.client.upload_object(
                        remote_path,
                        f,
                        hash_cb=hash_cb,
                        upload_cb=upload_cb,
                        **params)
                except Exception:
                    self._safe_progress_bar_finish(progress_bar)
                    raise
                finally:
                    self._safe_progress_bar_finish(progress_bar)
        print 'Upload completed'

    def main(self, local_path, container____path__=None):
        super(self.__class__, self)._run(container____path__)
        remote_path = self.path or path.basename(local_path)
        self._run(local_path=local_path, remote_path=remote_path)


@command(pithos_cmds)
class store_cat(_store_container_command):
    """Print remote file contents to console"""

    arguments = dict(
        range=RangeArgument('show range of data', '--range'),
        if_match=ValueArgument('show output if ETags match', '--if-match'),
        if_none_match=ValueArgument(
            'show output if ETags match',
            '--if-none-match'),
        if_modified_since=DateArgument(
            'show output modified since then',
            '--if-modified-since'),
        if_unmodified_since=DateArgument(
            'show output unmodified since then',
            '--if-unmodified-since'),
        object_version=ValueArgument(
            'get the specific version',
            '--object-version')
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        self.client.download_object(
            self.path,
            stdout,
            range_str=self['range'],
            version=self['object_version'],
            if_match=self['if_match'],
            if_none_match=self['if_none_match'],
            if_modified_since=self['if_modified_since'],
            if_unmodified_since=self['if_unmodified_since'])

    def main(self, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_download(_store_container_command):
    """Download remote object as local file
    If local destination is a directory:
    *   download <container>:<path> <local dir> -r
    will download all files on <container> prefixed as <path>,
    to <local dir>/<full path>
    *   download <container>:<path> <local dir> --exact-match
    will download only one file, exactly matching <path>
    ATTENTION: to download cont:dir1/dir2/file there must exist objects
    cont:dir1 and cont:dir1/dir2 of type application/directory
    To create directory objects, use /store mkdir
    """

    arguments = dict(
        resume=FlagArgument('Resume instead of overwrite', '--resume'),
        range=RangeArgument('show range of data', '--range'),
        if_match=ValueArgument('show output if ETags match', '--if-match'),
        if_none_match=ValueArgument(
            'show output if ETags match',
            '--if-none-match'),
        if_modified_since=DateArgument(
            'show output modified since then',
            '--if-modified-since'),
        if_unmodified_since=DateArgument(
            'show output unmodified since then',
            '--if-unmodified-since'),
        object_version=ValueArgument(
            'get the specific version',
            '--object-version'),
        poolsize=IntArgument('set pool size', '--with-pool-size'),
        progress_bar=ProgressBarArgument(
            'do not show progress bar',
            '--no-progress-bar',
            default=False),
        recursive=FlagArgument(
            'Download a remote path and all its contents',
            '--recursive')
    )

    @staticmethod
    def _is_dir(remote_dict):
        return 'application/directory' == remote_dict.get(
            'content_type',
            remote_dict.get('content-type', ''))

    def _outputs(self, local_path):
        """:returns: (local_file, remote_path)"""
        remotes = []
        if self['recursive']:
            r = self.client.container_get(
                prefix=self.path or '/',
                if_modified_since=self['if_modified_since'],
                if_unmodified_since=self['if_unmodified_since'])
            dirlist = dict()
            for remote in r.json:
                rname = remote['name'].strip('/')
                tmppath = ''
                for newdir in rname.strip('/').split('/')[:-1]:
                    tmppath = '/'.join([tmppath, newdir])
                    dirlist.update({tmppath.strip('/'): True})
                remotes.append((rname, store_download._is_dir(remote)))
            dir_remotes = [r[0] for r in remotes if r[1]]
            if not set(dirlist).issubset(dir_remotes):
                badguys = [bg.strip('/') for bg in set(
                    dirlist).difference(dir_remotes)]
                raiseCLIError(
                    'Some remote paths contain non existing directories',
                    details=['Missing remote directories:'] + badguys)
        elif self.path:
            r = self.client.get_object_info(
                self.path,
                version=self['object_version'])
            if store_download._is_dir(r):
                raiseCLIError(
                    'Illegal download: Remote object %s is a directory' % (
                        self.path),
                    details=['To download a directory, try --recursive'])
            if '/' in self.path.strip('/') and not local_path:
                raiseCLIError(
                    'Illegal download: remote object %s contains "/"' % (
                        self.path),
                    details=[
                        'To download an object containing "/" characters',
                        'either create the remote directories or',
                        'specify a non-directory local path for this object'])
            remotes = [(self.path, False)]
        if not remotes:
            if self.path:
                raiseCLIError(
                    'No matching path %s on container %s' % (
                        self.path,
                        self.container),
                    details=[
                        'To list the contents of %s, try:' % self.container,
                        '   /store list %s' % self.container])
            raiseCLIError(
                'Illegal download of container %s' % self.container,
                details=[
                    'To download a whole container, try:',
                    '   /store download --recursive <container>'])

        lprefix = path.abspath(local_path or path.curdir)
        if path.isdir(lprefix):
            for rpath, remote_is_dir in remotes:
                lpath = '/%s/%s' % (lprefix.strip('/'), rpath.strip('/'))
                if remote_is_dir:
                    if path.exists(lpath) and path.isdir(lpath):
                        continue
                    makedirs(lpath)
                elif path.exists(lpath):
                    if not self['resume']:
                        print('File %s exists, aborting...' % lpath)
                        continue
                    with open(lpath, 'rwb+') as f:
                        yield (f, rpath)
                else:
                    with open(lpath, 'wb+') as f:
                        yield (f, rpath)
        elif path.exists(lprefix):
            if len(remotes) > 1:
                raiseCLIError(
                    '%s remote objects cannot be merged in local file %s' % (
                        len(remotes),
                        local_path),
                    details=[
                        'To download multiple objects, local path should be',
                        'a directory, or use download without a local path'])
            (rpath, remote_is_dir) = remotes[0]
            if remote_is_dir:
                raiseCLIError(
                    'Remote directory %s should not replace local file %s' % (
                        rpath,
                        local_path))
            if self['resume']:
                with open(lprefix, 'rwb+') as f:
                    yield (f, rpath)
            else:
                raiseCLIError(
                    'Local file %s already exist' % local_path,
                    details=['Try --resume to overwrite it'])
        else:
            if len(remotes) > 1 or remotes[0][1]:
                raiseCLIError(
                    'Local directory %s does not exist' % local_path)
            with open(lprefix, 'wb+') as f:
                yield (f, remotes[0][0])

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    @errors.pithos.local_path
    def _run(self, local_path):
        #outputs = self._outputs(local_path)
        poolsize = self['poolsize']
        if poolsize:
            self.client.POOL_SIZE = int(poolsize)
        progress_bar = None
        try:
            for f, rpath in self._outputs(local_path):
                (
                    progress_bar,
                    download_cb) = self._safe_progress_bar(
                        'Download %s' % rpath)
                self.client.download_object(
                    rpath,
                    f,
                    download_cb=download_cb,
                    range_str=self['range'],
                    version=self['object_version'],
                    if_match=self['if_match'],
                    resume=self['resume'],
                    if_none_match=self['if_none_match'],
                    if_modified_since=self['if_modified_since'],
                    if_unmodified_since=self['if_unmodified_since'])
        except KeyboardInterrupt:
            from threading import enumerate as activethreads
            stdout.write('\nFinishing active threads ')
            for thread in activethreads():
                stdout.flush()
                try:
                    thread.join()
                    stdout.write('.')
                except RuntimeError:
                    continue
            print('\ndownload canceled by user')
            if local_path is not None:
                print('to resume, re-run with --resume')
        except Exception:
            self._safe_progress_bar_finish(progress_bar)
            raise
        finally:
            self._safe_progress_bar_finish(progress_bar)

    def main(self, container___path, local_path=None):
        super(self.__class__, self)._run(container___path)
        self._run(local_path=local_path)


@command(pithos_cmds)
class store_hashmap(_store_container_command):
    """Get the hash-map of an object"""

    arguments = dict(
        if_match=ValueArgument('show output if ETags match', '--if-match'),
        if_none_match=ValueArgument(
            'show output if ETags match',
            '--if-none-match'),
        if_modified_since=DateArgument(
            'show output modified since then',
            '--if-modified-since'),
        if_unmodified_since=DateArgument(
            'show output unmodified since then',
            '--if-unmodified-since'),
        object_version=ValueArgument(
            'get the specific version',
            '--object-version')
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        data = self.client.get_object_hashmap(
            self.path,
            version=self['object_version'],
            if_match=self['if_match'],
            if_none_match=self['if_none_match'],
            if_modified_since=self['if_modified_since'],
            if_unmodified_since=self['if_unmodified_since'])
        print_dict(data)

    def main(self, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_delete(_store_container_command):
    """Delete a container [or an object]
    How to delete a non-empty container:
    - empty the container:  /store delete -r <container>
    - delete it:            /store delete <container>
    .
    Semantics of directory deletion:
    .a preserve the contents: /store delete <container>:<directory>
    .    objects of the form dir/filename can exist with a dir object
    .b delete contents:       /store delete -r <container>:<directory>
    .    all dir/* objects are affected, even if dir does not exist
    .
    To restore a deleted object OBJ in a container CONT:
    - get object versions: /store versions CONT:OBJ
    .   and choose the version to be restored
    - restore the object:  /store copy --source-version=<version> CONT:OBJ OBJ
    """

    arguments = dict(
        until=DateArgument('remove history until that date', '--until'),
        yes=FlagArgument('Do not prompt for permission', '--yes'),
        recursive=FlagArgument(
            'empty dir or container and delete (if dir)',
            ('-r', '--recursive'))
    )

    def __init__(self, arguments={}):
        super(self.__class__, self).__init__(arguments)
        self['delimiter'] = DelimiterArgument(
            self,
            parsed_name='--delimiter',
            help='delete objects prefixed with <object><delimiter>')

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        if self.path:
            if self['yes'] or ask_user(
                    'Delete %s:%s ?' % (self.container, self.path)):
                self.client.del_object(
                    self.path,
                    until=self['until'],
                    delimiter=self['delimiter'])
            else:
                print('Aborted')
        else:
            if self['recursive']:
                ask_msg = 'Delete container contents'
            else:
                ask_msg = 'Delete container'
            if self['yes'] or ask_user('%s %s ?' % (ask_msg, self.container)):
                self.client.del_container(
                    until=self['until'],
                    delimiter=self['delimiter'])
            else:
                print('Aborted')

    def main(self, container____path__=None):
        super(self.__class__, self)._run(container____path__)
        self._run()


@command(pithos_cmds)
class store_purge(_store_container_command):
    """Delete a container and release related data blocks
    Non-empty containers can not purged.
    To purge a container with content:
    .   /store delete -r <container>
    .      objects are deleted, but data blocks remain on server
    .   /store purge <container>
    .      container and data blocks are released and deleted
    """

    arguments = dict(
        yes=FlagArgument('Do not prompt for permission', '--yes'),
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self):
        if self['yes'] or ask_user('Purge container %s?' % self.container):
                self.client.purge_container()
        else:
            print('Aborted')

    def main(self, container=None):
        super(self.__class__, self)._run(container)
        if container and self.container != container:
            raiseCLIError('Invalid container name %s' % container, details=[
                'Did you mean "%s" ?' % self.container,
                'Use --container for names containing :'])
        self._run()


@command(pithos_cmds)
class store_publish(_store_container_command):
    """Publish the object and print the public url"""

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        url = self.client.publish_object(self.path)
        print(url)

    def main(self, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_unpublish(_store_container_command):
    """Unpublish an object"""

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
            self.client.unpublish_object(self.path)

    def main(self, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_permissions(_store_container_command):
    """Get read and write permissions of an object
    Permissions are lists of users and user groups. There is read and write
    permissions. Users and groups with write permission have also read
    permission.
    """

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        r = self.client.get_object_sharing(self.path)
        print_dict(r)

    def main(self, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_setpermissions(_store_container_command):
    """Set permissions for an object
    New permissions overwrite existing permissions.
    Permission format:
    -   read=<username>[,usergroup[,...]]
    -   write=<username>[,usegroup[,...]]
    E.g. to give read permissions for file F to users A and B and write for C:
    .       /store setpermissions F read=A,B write=C
    """

    @errors.generic.all
    def format_permition_dict(self, permissions):
        read = False
        write = False
        for perms in permissions:
            splstr = perms.split('=')
            if 'read' == splstr[0]:
                read = [ug.strip() for ug in splstr[1].split(',')]
            elif 'write' == splstr[0]:
                write = [ug.strip() for ug in splstr[1].split(',')]
            else:
                msg = 'Usage:\tread=<groups,users> write=<groups,users>'
                raiseCLIError(None, msg)
        return (read, write)

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self, read, write):
        self.client.set_object_sharing(
            self.path,
            read_permition=read,
            write_permition=write)

    def main(self, container___path, *permissions):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        (read, write) = self.format_permition_dict(permissions)
        self._run(read, write)


@command(pithos_cmds)
class store_delpermissions(_store_container_command):
    """Delete all permissions set on object
    To modify permissions, use /store setpermssions
    """

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        self.client.del_object_sharing(self.path)

    def main(self, container___path):
        super(self.__class__, self)._run(
            container___path,
            path_is_optional=False)
        self._run()


@command(pithos_cmds)
class store_info(_store_container_command):
    """Get detailed information for user account, containers or objects
    to get account info:    /store info
    to get container info:  /store info <container>
    to get object info:     /store info <container>:<path>
    """

    arguments = dict(
        object_version=ValueArgument(
            'show specific version \ (applies only for objects)',
            '--object-version')
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        if self.container is None:
            r = self.client.get_account_info()
        elif self.path is None:
            r = self.client.get_container_info(self.container)
        else:
            r = self.client.get_object_info(
                self.path,
                version=self['object_version'])
        print_dict(r)

    def main(self, container____path__=None):
        super(self.__class__, self)._run(container____path__)
        self._run()


@command(pithos_cmds)
class store_meta(_store_container_command):
    """Get metadata for account, containers or objects"""

    arguments = dict(
        detail=FlagArgument('show detailed output', '-l'),
        until=DateArgument('show metadata until then', '--until'),
        object_version=ValueArgument(
            'show specific version \ (applies only for objects)',
            '--object-version')
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        until = self['until']
        if self.container is None:
            if self['detail']:
                r = self.client.get_account_info(until=until)
            else:
                r = self.client.get_account_meta(until=until)
                r = pretty_keys(r, '-')
            if r:
                print(bold(self.client.account))
        elif self.path is None:
            if self['detail']:
                r = self.client.get_container_info(until=until)
            else:
                cmeta = self.client.get_container_meta(until=until)
                ometa = self.client.get_container_object_meta(until=until)
                r = {}
                if cmeta:
                    r['container-meta'] = pretty_keys(cmeta, '-')
                if ometa:
                    r['object-meta'] = pretty_keys(ometa, '-')
        else:
            if self['detail']:
                r = self.client.get_object_info(
                    self.path,
                    version=self['object_version'])
            else:
                r = self.client.get_object_meta(
                    self.path,
                    version=self['object_version'])
            if r:
                r = pretty_keys(pretty_keys(r, '-'))
        if r:
            print_dict(r)

    def main(self, container____path__=None):
        super(self.__class__, self)._run(container____path__)
        self._run()


@command(pithos_cmds)
class store_setmeta(_store_container_command):
    """Set a piece of metadata for account, container or object
    Metadata are formed as key:value pairs
    """

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self, metakey, metaval):
        if not self.container:
            self.client.set_account_meta({metakey: metaval})
        elif not self.path:
            self.client.set_container_meta({metakey: metaval})
        else:
            self.client.set_object_meta(self.path, {metakey: metaval})

    def main(self, metakey, metaval, container____path__=None):
        super(self.__class__, self)._run(container____path__)
        self._run(metakey=metakey, metaval=metaval)


@command(pithos_cmds)
class store_delmeta(_store_container_command):
    """Delete metadata with given key from account, container or object
    Metadata are formed as key:value objects
    - to get metadata of current account:     /store meta
    - to get metadata of a container:         /store meta <container>
    - to get metadata of an object:           /store meta <container>:<path>
    """

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self, metakey):
        if self.container is None:
            self.client.del_account_meta(metakey)
        elif self.path is None:
            self.client.del_container_meta(metakey)
        else:
            self.client.del_object_meta(self.path, metakey)

    def main(self, metakey, container____path__=None):
        super(self.__class__, self)._run(container____path__)
        self._run(metakey)


@command(pithos_cmds)
class store_quota(_store_account_command):
    """Get quota for account or container"""

    arguments = dict(
        in_bytes=FlagArgument('Show result in bytes', ('-b', '--bytes'))
    )

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self):
        if self.container:
            reply = self.client.get_container_quota(self.container)
        else:
            reply = self.client.get_account_quota()
        if not self['in_bytes']:
            for k in reply:
                reply[k] = format_size(reply[k])
        print_dict(pretty_keys(reply, '-'))

    def main(self, container=None):
        super(self.__class__, self)._run()
        self.container = container
        self._run()


@command(pithos_cmds)
class store_setquota(_store_account_command):
    """Set new quota for account or container
    By default, quota is set in bytes
    Users may specify a different unit, e.g:
    /store setquota 2.3GB mycontainer
    Accepted units: B, KiB (1024 B), KB (1000 B), MiB, MB, GiB, GB, TiB, TB
    """

    @errors.generic.all
    def _calculate_quota(self, user_input):
        quota = 0
        try:
            quota = int(user_input)
        except ValueError:
            index = 0
            digits = [str(num) for num in range(0, 10)] + ['.']
            while user_input[index] in digits:
                index += 1
            quota = user_input[:index]
            format = user_input[index:]
            try:
                return to_bytes(quota, format)
            except Exception as qe:
                msg = 'Failed to convert %s to bytes' % user_input,
                raiseCLIError(qe, msg, details=[
                    'Syntax: setquota <quota>[format] [container]',
                    'e.g.: setquota 2.3GB mycontainer',
                    'Acceptable formats:',
                    '(*1024): B, KiB, MiB, GiB, TiB',
                    '(*1000): B, KB, MB, GB, TB'])
        return quota

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self, quota):
        if self.container:
            self.client.container = self.container
            self.client.set_container_quota(quota)
        else:
            self.client.set_account_quota(quota)

    def main(self, quota, container=None):
        super(self.__class__, self)._run()
        quota = self._calculate_quota(quota)
        self.container = container
        self._run(quota)


@command(pithos_cmds)
class store_versioning(_store_account_command):
    """Get  versioning for account or container"""

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self):
        if self.container:
            r = self.client.get_container_versioning(self.container)
        else:
            r = self.client.get_account_versioning()
        print_dict(r)

    def main(self, container=None):
        super(self.__class__, self)._run()
        self.container = container
        self._run()


@command(pithos_cmds)
class store_setversioning(_store_account_command):
    """Set versioning mode (auto, none) for account or container"""

    def _check_versioning(self, versioning):
        if versioning and versioning.lower() in ('auto', 'none'):
            return versioning.lower()
        raiseCLIError('Invalid versioning %s' % versioning, details=[
            'Versioning can be auto or none'])

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    def _run(self, versioning):
        if self.container:
            self.client.container = self.container
            self.client.set_container_versioning(versioning)
        else:
            self.client.set_account_versioning(versioning)

    def main(self, versioning, container=None):
        super(self.__class__, self)._run()
        self._run(self._check_versioning(versioning))


@command(pithos_cmds)
class store_group(_store_account_command):
    """Get groups and group members"""

    @errors.generic.all
    @errors.pithos.connection
    def _run(self):
        r = self.client.get_account_group()
        print_dict(pretty_keys(r, '-'))

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(pithos_cmds)
class store_setgroup(_store_account_command):
    """Set a user group"""

    @errors.generic.all
    @errors.pithos.connection
    def _run(self, groupname, *users):
        self.client.set_account_group(groupname, users)

    def main(self, groupname, *users):
        super(self.__class__, self)._run()
        if users:
            self._run(groupname, *users)
        else:
            raiseCLIError('No users to add in group %s' % groupname)


@command(pithos_cmds)
class store_delgroup(_store_account_command):
    """Delete a user group"""

    @errors.generic.all
    @errors.pithos.connection
    def _run(self, groupname):
        self.client.del_account_group(groupname)

    def main(self, groupname):
        super(self.__class__, self)._run()
        self._run(groupname)


@command(pithos_cmds)
class store_sharers(_store_account_command):
    """List the accounts that share objects with current user"""

    arguments = dict(
        detail=FlagArgument('show detailed output', '-l'),
        marker=ValueArgument('show output greater then marker', '--marker')
    )

    @errors.generic.all
    @errors.pithos.connection
    def _run(self):
        accounts = self.client.get_sharing_accounts(marker=self['marker'])
        if self['detail']:
            print_items(accounts)
        else:
            print_items([acc['name'] for acc in accounts])

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(pithos_cmds)
class store_versions(_store_container_command):
    """Get the list of object versions
    Deleted objects may still have versions that can be used to restore it and
    get information about its previous state.
    The version number can be used in a number of other commands, like info,
    copy, move, meta. See these commands for more information, e.g.
    /store info -h
    """

    @errors.generic.all
    @errors.pithos.connection
    @errors.pithos.container
    @errors.pithos.object_path
    def _run(self):
        versions = self.client.get_object_versionlist(self.path)
        print_items([dict(id=vitem[0], created=strftime(
            '%d-%m-%Y %H:%M:%S',
            localtime(float(vitem[1])))) for vitem in versions])

    def main(self, container___path):
        super(store_versions, self)._run(
            container___path,
            path_is_optional=False)
        self._run()



# -*- coding: utf-8 -*-
#
# Copyright 2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.


from kamaki.clients.commissioning import (
    CanonifyException,
    SpecifyException,
    Specificator,
    Null,
    Integer,
    Text,
    Tuple,
    ListOf,
    Dict,
    Args)
from random import choice, randint

Context = Dict(classname='Context', null=True, show=False)


class Name(Text):
    def init(self):
        self.opts.update({'regex': "[\w.:@+/-]+", 'maxlen': 512})
        Text.init(self)

    def _random_choice(self, kw):
        alphabet = u'abcdef_1233490.:@/-αβγδεζ'
        length = randint(1, 48)
        return ''.join(choice(alphabet) for _ in xrange(length))


class Nonnegative(Integer):
    def init(self):
        self.opts.update({'minimum': 0})


class Positive(Integer):
    def init(self):
        self.opts.update({'minimum': 1})


QH_PRACTICALLY_INFINITE = 10 ** 32

Serial = Positive(classname='Serial')

ClientKey = Name(classname='ClientKey')
Nothing = Null(classname='Nothing')

Entity = Name(classname='Entity')
Owner = Name(classname='Owner')
Key = Text(classname='Key')
NewKey = Text(classname='Newkey')
OwnerKey = Text(classname='OwnerKey')
Resource = Name(classname='Resource')
Policy = Name(classname='Policy')

Quantity = Integer(classname='Quantity')
Capacity = Nonnegative(classname='Capacity')
ImportLimit = Nonnegative(classname='ImportLimit')
ExportLimit = Nonnegative(classname='ExportLimit')
QuantityDelta = Integer(classname='QuantityDelta')
CapacityDelta = Integer(classname='CapacityDelta')
ImportLimitDelta = Integer(classname='ImportLimitDelta')
ExportLimitDelta = Integer(classname='ExportLimitDelta')
Imported = Nonnegative(classname='Imported')
Exported = Nonnegative(classname='Exported')
Returned = Nonnegative(classname='Returned')
Released = Nonnegative(classname='Released')
Flags = Nonnegative(classname='Flags')
Index = Nonnegative(classname='Index')

Timepoint = Text(classname='Timepoint', maxlen=24)
Reason = Text(classname='Reason', regex='(ACCEPT|REJECT):.*', maxlen=128)


class QuotaholderAPI(Specificator):

    def create_entity(self, context=Context, create_entity=ListOf(
            Entity, Owner, Key, OwnerKey, nonempty=1)):
        rejected = ListOf(Index)
        return rejected

    def set_entity_key(
        self,
        context=Context,
        set_entity_key=ListOf(Entity, Key, NewKey)
    ):
        rejected = ListOf(Entity)
        return rejected

    def list_entities(
        self,
        context=Context,
        entity=Entity,
        key=Key
    ):
        entities = ListOf(Entity)
        return entities

    def get_entity(
        self,
        context=Context,
        get_entity=ListOf(Entity, Key, nonempty=1)
    ):
        entities = ListOf(Entity, Owner)
        return entities

    def get_limits(
        self,
        context=Context,
        get_limits=ListOf(Policy, nonempty=1)
    ):
        limits = ListOf(Policy, Quantity, Capacity,
                        ImportLimit, ExportLimit)
        return limits

    def set_limits(
        self,
        context=Context,
        set_limits=ListOf(
            Policy,
            Quantity,
            Capacity,
            ImportLimit,
            ExportLimit,
            nonempty=1)
    ):
        rejected = ListOf(Policy)
        return rejected

    def get_holding(
        self,
        context=Context,
        get_holding=ListOf(Entity, Resource, Key)
    ):
        holdings = ListOf(
            Entity,
            Resource,
            Policy,
            Imported,
            Exported,
            Returned,
            Released,
            Flags)
        return holdings

    def set_holding(
        self,
        context=Context,
        set_holding=ListOf(Entity, Resource, Key, Policy, Flags)
    ):
        rejected = ListOf(Entity, Resource, Policy)
        return rejected

    def init_holding(
        self,
        context=Context,
        init_holding=ListOf(
            Entity,
            Resource,
            Key,
            Policy,
            Imported,
            Exported,
            Returned,
            Released,
            Flags)
    ):
        rejected = ListOf(Index)
        return rejected

    def reset_holding(
        self,
        context=Context,
        reset_holding=ListOf(
            Entity,
            Resource,
            Key,
            Imported,
            Exported,
            Returned,
            Released)
    ):
        rejected = ListOf(Index)
        return rejected

    def release_holding(
        self,
        context=Context,
        release_holding=ListOf(Entity, Resource, Key)
    ):
        rejected = ListOf(Index)
        return rejected

    def list_resources(self, context=Context, entity=Entity, key=Key):
        resources = ListOf(Resource)
        return resources

    def list_holdings(
        self,
        context=Context,
        list_holdings=ListOf(Entity, Key)
    ):
        rejected = ListOf(Entity)
        holdings_list = ListOf(ListOf(Entity, Resource,
                                      Imported, Exported,
                                      Returned, Released))
        return Tuple(holdings_list, rejected)

    def get_quota(
        self,
        context=Context,
        get_quota=ListOf(Entity, Resource, Key)
    ):
        quotas = ListOf(Entity, Resource,
                        Quantity, Capacity,
                        ImportLimit, ExportLimit,
                        Imported, Exported,
                        Returned, Released,
                        Flags)
        return quotas

    def set_quota(
        self,
        context=Context,
        set_quota=ListOf(
            Entity,
            Resource,
            Key,
            Quantity,
            Capacity,
            ImportLimit,
            ExportLimit,
            Flags)
    ):
        rejected = ListOf(Entity, Resource)
        return rejected

    def add_quota(
        self,
        context=Context,
        clientkey=ClientKey,
        serial=Serial,
        sub_quota=ListOf(
            Entity,
            Resource,
            Key,
            QuantityDelta,
            CapacityDelta,
            ImportLimitDelta,
            ExportLimitDelta),
        add_quota=ListOf(
            Entity,
            Resource,
            Key,
            QuantityDelta,
            CapacityDelta,
            ImportLimitDelta,
            ExportLimitDelta)
    ):
        rejected = ListOf(Entity, Resource)
        return rejected

    def query_serials(
        self,
        context=Context,
        clientkey=ClientKey,
        serials=ListOf(Serial)
    ):
        return ListOf(Serial)

    def ack_serials(
        self,
        context=Context,
        clientkey=ClientKey,
        serials=ListOf(Serial)
    ):
        return Nothing

    def issue_commission(
        self,
        context=Context,
        target=Entity,
        key=Key,
        clientkey=ClientKey,
        name=Text(default=''),
        provisions=ListOf(Entity, Resource, Quantity)
    ):
        return Serial

    def accept_commission(
        self,
        context=Context,
        clientkey=ClientKey,
        serials=ListOf(Serial),
        reason=Text(default='ACCEPT')
    ):
        return Nothing

    def reject_commission(
        self,
        context=Context,
        clientkey=ClientKey,
        serials=ListOf(Serial),
        reason=Text(default='REJECT')
    ):
        return Nothing

    def get_pending_commissions(
        self,
        context=Context,
        clientkey=ClientKey
    ):
        pending = ListOf(Serial)
        return pending

    def resolve_pending_commissions(
        self,
        context=Context,
        clientkey=ClientKey,
        max_serial=Serial,
        accept_set=ListOf(Serial)
    ):
        return Nothing

    def release_entity(
        self,
        context=Context,
        release_entity=ListOf(Entity, Key, nonempty=1)
    ):
        rejected = ListOf(Entity)
        return rejected

    def get_timeline(
        self,
        context=Context,
        after=Timepoint,
        before=Timepoint,
        get_timeline=ListOf(Entity, Resource, Key)
    ):
        timeline = ListOf(Dict(
            serial=Serial,
            source=Entity,
            target=Entity,
            resource=Resource,
            name=Name(),
            quantity=Quantity,
            source_allocated=Quantity,
            source_allocated_through=Quantity,
            source_inbound=Quantity,
            source_inbound_through=Quantity,
            source_outbound=Quantity,
            source_outbound_through=Quantity,
            target_allocated=Quantity,
            target_allocated_through=Quantity,
            target_inbound=Quantity,
            target_inbound_through=Quantity,
            target_outbound=Quantity,
            target_outbound_through=Quantity,
            issue_time=Timepoint,
            log_time=Timepoint,
            reason=Reason,
            strict=True))
        return timeline



# Copyright 2012-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

import logging
from sys import argv, exit, stdout
from os.path import basename
from inspect import getargspec

from kamaki.cli.argument import ArgumentParseManager
from kamaki.cli.history import History
from kamaki.cli.utils import print_dict, red, magenta, yellow
from kamaki.cli.errors import CLIError

_help = False
_debug = False
_include = False
_verbose = False
_colors = False
kloger = None

#  command auxiliary methods

_best_match = []


def _arg2syntax(arg):
    return arg.replace(
        '____', '[:').replace(
            '___', ':').replace(
                '__', ']').replace(
                    '_', ' ')


def _construct_command_syntax(cls):
        spec = getargspec(cls.main.im_func)
        args = spec.args[1:]
        n = len(args) - len(spec.defaults or ())
        required = ' '.join(['<%s>' % _arg2syntax(x) for x in args[:n]])
        optional = ' '.join(['[%s]' % _arg2syntax(x) for x in args[n:]])
        cls.syntax = ' '.join(x for x in [required, optional] if x)
        if spec.varargs:
            cls.syntax += ' <%s ...>' % spec.varargs


def _num_of_matching_terms(basic_list, attack_list):
    if not attack_list:
        return len(basic_list)

    matching_terms = 0
    for i, term in enumerate(basic_list):
        try:
            if term != attack_list[i]:
                break
        except IndexError:
            break
        matching_terms += 1
    return matching_terms


def _update_best_match(name_terms, prefix=[]):
    if prefix:
        pref_list = prefix if isinstance(prefix, list) else prefix.split('_')
    else:
        pref_list = []

    num_of_matching_terms = _num_of_matching_terms(name_terms, pref_list)
    global _best_match
    if not prefix:
        _best_match = []

    if num_of_matching_terms and len(_best_match) <= num_of_matching_terms:
        if len(_best_match) < num_of_matching_terms:
            _best_match = name_terms[:num_of_matching_terms]
        return True
    return False


def command(cmd_tree, prefix='', descedants_depth=1):
    """Load a class as a command
        e.g. spec_cmd0_cmd1 will be command spec cmd0

        :param cmd_tree: is initialized in cmd_spec file and is the structure
            where commands are loaded. Var name should be _commands
        :param prefix: if given, load only commands prefixed with prefix,
        :param descedants_depth: is the depth of the tree descedants of the
            prefix command. It is used ONLY if prefix and if prefix is not
            a terminal command

        :returns: the specified class object
    """

    def wrap(cls):
        global kloger
        cls_name = cls.__name__

        if not cmd_tree:
            if _debug:
                kloger.warning('command %s found but not loaded' % cls_name)
            return cls

        name_terms = cls_name.split('_')
        if not _update_best_match(name_terms, prefix):
            if _debug:
                kloger.warning('%s failed to update_best_match' % cls_name)
            return None

        global _best_match
        max_len = len(_best_match) + descedants_depth
        if len(name_terms) > max_len:
            partial = '_'.join(name_terms[:max_len])
            if not cmd_tree.has_command(partial):  # add partial path
                cmd_tree.add_command(partial)
            if _debug:
                kloger.warning('%s failed max_len test' % cls_name)
            return None

        (
            cls.description, sep, cls.long_description
        ) = cls.__doc__.partition('\n')
        _construct_command_syntax(cls)

        cmd_tree.add_command(cls_name, cls.description, cls)
        return cls
    return wrap


cmd_spec_locations = [
    'kamaki.cli.commands',
    'kamaki.commands',
    'kamaki.cli',
    'kamaki',
    '']


#  Generic init auxiliary functions


def _setup_logging(silent=False, debug=False, verbose=False, include=False):
    """handle logging for clients package"""

    def add_handler(name, level, prefix=''):
        h = logging.StreamHandler()
        fmt = logging.Formatter(prefix + '%(message)s')
        h.setFormatter(fmt)
        logger = logging.getLogger(name)
        logger.addHandler(h)
        logger.setLevel(level)

    if silent:
        add_handler('', logging.CRITICAL)
        return

    if debug:
        add_handler('requests', logging.INFO, prefix='* ')
        add_handler('clients.send', logging.DEBUG, prefix='> ')
        add_handler('clients.recv', logging.DEBUG, prefix='< ')
        add_handler('kamaki', logging.DEBUG, prefix='(debug): ')
    elif verbose:
        add_handler('requests', logging.INFO, prefix='* ')
        add_handler('clients.send', logging.INFO, prefix='> ')
        add_handler('clients.recv', logging.INFO, prefix='< ')
        add_handler('kamaki', logging.INFO, prefix='(i): ')
    if include:
        add_handler('data.send', logging.INFO, prefix='>[data]: ')
        add_handler('data.recv', logging.INFO, prefix='<[data]: ')
    add_handler('kamaki', logging.WARNING, prefix='(warning): ')
    global kloger
    kloger = logging.getLogger('kamaki')


def _init_session(arguments):
    global _help
    _help = arguments['help'].value
    global _debug
    _debug = arguments['debug'].value
    global _include
    _include = arguments['include'].value
    global _verbose
    _verbose = arguments['verbose'].value
    global _colors
    _colors = arguments['config'].get('global', 'colors')
    if not (stdout.isatty() and _colors == 'on'):
        from kamaki.cli.utils import remove_colors
        remove_colors()
    _silent = arguments['silent'].value
    _setup_logging(_silent, _debug, _verbose, _include)


def _load_spec_module(spec, arguments, module):
    spec_name = arguments['config'].get(spec, 'cli')
    if spec_name is None:
        return None
    pkg = None
    for location in cmd_spec_locations:
        location += spec_name if location == '' else '.%s' % spec_name
        try:
            pkg = __import__(location, fromlist=[module])
            return pkg
        except ImportError:
            continue
    return pkg


def _groups_help(arguments):
    global _debug
    global kloger
    descriptions = {}
    for spec in arguments['config'].get_groups():
        pkg = _load_spec_module(spec, arguments, '_commands')
        if pkg:
            cmds = None
            try:
                _cnf = arguments['config']
                cmds = [cmd for cmd in getattr(pkg, '_commands') if _cnf.get(
                    cmd.name, 'cli')]
            except AttributeError:
                if _debug:
                    kloger.warning('No description for %s' % spec)
            try:
                for cmd in cmds:
                    descriptions[cmd.name] = cmd.description
            except TypeError:
                if _debug:
                    kloger.warning('no cmd specs in module %s' % spec)
        elif _debug:
            kloger.warning('Loading of %s cmd spec failed' % spec)
    print('\nOptions:\n - - - -')
    print_dict(descriptions)


def _load_all_commands(cmd_tree, arguments):
    _cnf = arguments['config']
    specs = [spec for spec in _cnf.get_groups() if _cnf.get(spec, 'cli')]
    for spec in specs:
        try:
            spec_module = _load_spec_module(spec, arguments, '_commands')
            spec_commands = getattr(spec_module, '_commands')
        except AttributeError:
            if _debug:
                global kloger
                kloger.warning('No valid description for %s' % spec)
            continue
        for spec_tree in spec_commands:
            if spec_tree.name == spec:
                cmd_tree.add_tree(spec_tree)
                break


#  Methods to be used by CLI implementations


def print_subcommands_help(cmd):
    printout = {}
    for subcmd in cmd.get_subcommands():
        spec, sep, print_path = subcmd.path.partition('_')
        printout[print_path.replace('_', ' ')] = subcmd.description
    if printout:
        print('\nOptions:\n - - - -')
        print_dict(printout)


def update_parser_help(parser, cmd):
    global _best_match
    parser.syntax = parser.syntax.split('<')[0]
    parser.syntax += ' '.join(_best_match)

    description = ''
    if cmd.is_command:
        cls = cmd.get_class()
        parser.syntax += ' ' + cls.syntax
        parser.update_arguments(cls().arguments)
        description = getattr(cls, 'long_description', '')
        description = description.strip()
    else:
        parser.syntax += ' <...>'
    if cmd.has_description:
        parser.parser.description = cmd.help + (
            ('\n%s' % description) if description else '')
    else:
        parser.parser.description = description


def print_error_message(cli_err):
    errmsg = '%s' % cli_err
    if cli_err.importance == 1:
        errmsg = magenta(errmsg)
    elif cli_err.importance == 2:
        errmsg = yellow(errmsg)
    elif cli_err.importance > 2:
        errmsg = red(errmsg)
    stdout.write(errmsg)
    for errmsg in cli_err.details:
        print('| %s' % errmsg)


def exec_cmd(instance, cmd_args, help_method):
    try:
        return instance.main(*cmd_args)
    except TypeError as err:
        if err.args and err.args[0].startswith('main()'):
            print(magenta('Syntax error'))
            if _debug:
                raise err
            if _verbose:
                print(unicode(err))
            help_method()
        else:
            raise
    return 1


def get_command_group(unparsed, arguments):
    groups = arguments['config'].get_groups()
    for term in unparsed:
        if term.startswith('-'):
            continue
        if term in groups:
            unparsed.remove(term)
            return term
        return None
    return None


def set_command_params(parameters):
    """Add a parameters list to a command

    :param paramters: (list of str) a list of parameters
    """
    global command
    def_params = list(command.func_defaults)
    def_params[0] = parameters
    command.func_defaults = tuple(def_params)


#  CLI Choice:

def run_one_cmd(exe_string, parser):
    global _history
    _history = History(
        parser.arguments['config'].get('history', 'file'))
    _history.add(' '.join([exe_string] + argv[1:]))
    from kamaki.cli import one_command
    one_command.run(parser, _help)


def run_shell(exe_string, parser):
    from command_shell import _init_shell
    shell = _init_shell(exe_string, parser)
    _load_all_commands(shell.cmd_tree, parser.arguments)
    shell.run(parser)


def main():
    try:
        exe = basename(argv[0])
        parser = ArgumentParseManager(exe)

        if parser.arguments['version'].value:
            exit(0)

        _init_session(parser.arguments)

        from kamaki.cli.utils import suggest_missing
        suggest_missing()

        if parser.unparsed:
            run_one_cmd(exe, parser)
        elif _help:
            parser.parser.print_help()
            _groups_help(parser.arguments)
        else:
            run_shell(exe, parser)
    except CLIError as err:
        print_error_message(err)
        if _debug:
            raise err
        exit(1)
    except Exception as er:
        print('Unknown Error: %s' % er)
        if _debug:
            raise
        exit(1)



__version__ = "0.7.8"
__version_info__ = ['0', '7', '8']
__version_vcs_info__ = {'branch': 'master',
 'revid': '4d783ed',
 'revno': 867,
 'toplevel': '/home/saxtouri/src/kamaki/kamaki'}
__version_user_info__ = "saxtouri@terminus"



# Copyright 2011-2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from kamaki.cli import command
from kamaki.clients.astakos import AstakosClient
from kamaki.cli.utils import print_dict
from kamaki.cli.commands import _command_init, errors
from kamaki.cli.command_tree import CommandTree
# from kamaki.cli.argument import ValueArgument

astakos_cmds = CommandTree('astakos', 'Astakos API commands')
_commands = [astakos_cmds]


class _astakos_init(_command_init):

    @errors.generic.all
    @errors.astakos.load
    def _run(self):
        token = self.config.get('astakos', 'token')\
            or self.config.get('global', 'token')
        base_url = self.config.get('astakos', 'url')\
            or self.config.get('global', 'url')
        self.client = AstakosClient(base_url=base_url, token=token)

    def main(self):
        self._run


@command(astakos_cmds)
class astakos_authenticate(_astakos_init):
    """Authenticate a user
    Get user information (e.g. unique account name) from token
    Token should be set in settings:
    *  check if a token is set    /config get token
    *  permanently set a token    /config set token <token>
    Token can also be provided as a parameter
    """

    @errors.generic.all
    @errors.astakos.authenticate
    def _run(self, custom_token=None):
        super(self.__class__, self)._run()
        reply = self.client.authenticate(custom_token)
        print_dict(reply)

    def main(self, custom_token=None):
        self._run(custom_token)
